"""Convert CSV of model outputs to Stony Brook BMI formats.

Output directory tree for single class outputs:
├── heatmap_jsons
│   ├── heatmap-SLIDEID.json
│   └── meta-SLIDEID.json
└── heatmap_txt
    ├── color-SLIDEID
    └── prediction-SLIDEID

Output directory tree for multi-class outputs:
├── heatmap_jsons
│   └── CLASS_LABEL
│       ├── heatmap-SLIDEID.json
│       └── meta-SLIDEID.json
└── heatmap_txt
    └── CLASS_LABEL
        ├── color-SLIDEID
        └── prediction-SLIDEID
"""

from __future__ import annotations

import json
import multiprocessing
import pprint
import random
import shutil
import time
from pathlib import Path
from typing import Any

import click
import numpy as np
import numpy.typing as npt
import pandas as pd
import tqdm

from ..wsi import WSI
from ..wsi import CanReadRegion


def _box_to_polygon(
    *, minx: float, miny: float, width: float, height: float
) -> list[tuple[float, float]]:
    """Get coordinates of a box polygon."""
    maxx = minx + width
    maxy = miny + height
    return [(maxx, miny), (maxx, maxy), (minx, maxy), (minx, miny), (maxx, miny)]


def write_heatmap_and_meta_json_lines(
    input: str | Path,
    output_heatmap: str | Path,
    output_meta: str | Path,
    slide_width: int,
    slide_height: int,
    execution_id: str,
    study_id: str,
    case_id: str,
    subject_id: str,
    class_name: str,
    run_metadata: dict,
) -> None:
    """Write JSON-lines files for one slide."""

    # Run this before defining the function so the entire JSON file has the same
    # execution time value.
    date = int(time.time())

    version_dict: dict = run_metadata["runtime"]["git"] or {}
    version_dict["model_path"] = run_metadata["model_weights"]["weights_file"]
    version_dict["model_hash"] = run_metadata["model_weights"]["weights_sha256"]
    version_dict["model_url"] = run_metadata["model_weights"]["weights_url"]
    version_dict["model_ver"] = None

    def row_to_json(row: pd.Series) -> dict[str, Any]:
        minx, miny, width, height = (
            row["minx"],
            row["miny"],
            row["width"],
            row["height"],
        )
        patch_area_base_pixels = width * height
        # Scale the values to be a ratio of the whole slide dimensions. All of the
        # values in the dictionary (except 'footprint') use these normalized
        # coordinates.
        minx = float(minx / slide_width)
        miny = float(miny / slide_height)
        width = float(width / slide_width)
        height = float(height / slide_height)
        maxx = minx + width
        maxy = miny + height
        centerx = (minx + maxx) / 2
        centery = (miny + maxy) / 2
        coords = _box_to_polygon(minx=minx, miny=miny, width=width, height=height)

        # Get the probabilites from the model.
        if f"prob_{class_name}" not in row.index:
            raise KeyError(f"class name not found in results: {class_name}")
        class_probability: float = row[f"prob_{class_name}"]
        return {
            "type": "Feature",
            "parent_id": "self",
            "object_type": "heatmap_multiple",
            "x": centerx,
            "y": centery,
            "normalized": "true",
            "footprint": patch_area_base_pixels,
            "geometry": {
                "coordinates": [coords],
                "type": "Polygon",
            },
            "provenance": {
                "analysis": {
                    "source": "computer",
                    "execution_id": execution_id,
                    "cancer_type": "quip",
                    "study_id": study_id,
                    "computation": "heatmap",
                    "execution_time": run_metadata["timestamp"],
                },
                "image": {
                    "case_id": case_id,
                    "subject_id": subject_id,
                },
                "version": version_dict,
            },
            "bbox": [minx, miny, maxx, maxy],
            "properties": {
                "multiheat_param": {
                    "human_weight": -1,
                    "metric_array": [class_probability],
                    "heatname_array": [class_name],
                    "weight_array": ["1"],
                },
                "metric_value": class_probability,
                "metric_type": "tile_dice",
                "human_mark": -1,
            },
            "date": {"$date": date},
        }

    # Write heatmap JSON lines file.
    df = pd.read_csv(input)
    features = df.apply(row_to_json, axis=1)
    features = features.tolist()
    features = (json.dumps(row) for row in features)
    with open(output_heatmap, "w") as f:
        f.writelines(line + "\n" for line in features)

    # Write meta file.
    meta_dict = {
        "color": "yellow",  # This is copied from the lung cancer detection code.
        "title": execution_id,
        "image": {
            "case_id": case_id,
            "subject_id": subject_id,
        },
        "provenance": {
            "analysis_execution_id": execution_id,
            "analysis_execution_date": run_metadata["timestamp"],
            "study_id": study_id,
            "type": "computer",
            "version": version_dict,
        },
        "submit_date": {"$date": date},
        "randval": random.uniform(0, 1),
    }
    with open(output_meta, "w") as f:
        json.dump(meta_dict, f)


def write_heatmap_txt(
    input: str | Path, output: str | Path, class_names: list[str]
) -> None:
    df = pd.read_csv(input)
    # TODO: should we round and cast to int here?
    df.loc[:, "x_loc"] = (df.minx + (df.width / 2)).round().astype(int)
    df.loc[:, "y_loc"] = (df.miny + (df.height / 2)).round().astype(int)
    prob_cols = [f"prob_{c}" for c in class_names]
    cols = ["x_loc", "y_loc", *prob_cols]
    df = df.loc[:, cols]
    # Rename 'prob_NAME' columns to 'NAME'.
    df = df.rename(columns={c: c[5:] for c in prob_cols})
    df.to_csv(output, index=False, sep=" ")


def write_color_txt(
    input: str | Path,
    output: str | Path,
    slide: CanReadRegion,
    num_processes: int = 6,
) -> None:
    def whiteness(arr: npt.ArrayLike) -> float:
        arr = np.asarray(arr)
        return np.std(arr, axis=(0, 1)).mean()  # type: ignore

    def blackness(arr: npt.ArrayLike) -> float:
        arr = np.asarray(arr)
        return arr.mean()  # type: ignore

    def redness(arr: npt.ArrayLike) -> float:
        arr = np.asarray(arr)
        r, g, b = arr[:, :, 0], arr[:, :, 1], arr[:, :, 2]
        # boolean multiplication is logical and
        return np.mean((r >= 190) * (g <= 100) * (b <= 100))  # type: ignore

    global get_color  # Hack to please multiprocessing.

    def get_color(row: pd.Series) -> tuple[float, float, float]:
        patch_im = slide.read_region(
            location=(row["minx"], row["miny"]),
            level=0,
            size=(row["width"], row["height"]),
        )
        arr = np.asarray(patch_im)
        white = whiteness(arr)
        black = blackness(arr)
        red = redness(arr)
        return white, black, red

    df = pd.read_csv(input)
    df_rows_as_dicts = df.to_dict("records")

    # https://stackoverflow.com/a/45276885/5666087
    with multiprocessing.Pool(num_processes) as p:
        results = list(
            tqdm.tqdm(
                p.imap(get_color, df_rows_as_dicts, chunksize=16),  # type: ignore
                total=len(df_rows_as_dicts),
            )
        )
    df.loc[:, ["whiteness", "blackness", "redness"]] = results
    cx = df.minx + ((df.minx + df.width) / 2)
    assert np.array_equal(cx.astype(int), cx), "not all center x's are integer"
    df.loc[:, "cx"] = cx.astype(int)
    del cx  # sanity

    cy = df.miny + ((df.miny + df.height) / 2)
    assert np.array_equal(cy.astype(int), cy), "not all center y's are integer"
    df.loc[:, "cy"] = cy.astype(int)
    del cy

    df = df.loc[:, ["cx", "cy", "whiteness", "blackness", "redness"]]
    df.to_csv(output, header=False, index=False, sep=" ")


@click.command()
@click.argument(
    "results_dir",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, path_type=Path, resolve_path=True
    ),
)
@click.argument(
    "output",
    type=click.Path(exists=False, path_type=Path, resolve_path=True),
)
@click.option(
    "--wsi-dir",
    required=True,
    help="Directory with whole slide images.",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, path_type=Path, resolve_path=True
    ),
)
@click.option("--execution-id", required=True, help="Unique execution ID for this run.")
@click.option("--study-id", required=True, help="Study ID, like TCGA-BRCA.")
@click.option(
    "--make-color-text/--no-make-color-text",
    default=False,
    help="Make text files with color information for each patch. NOTE: this can add"
    " several minutes of processing time per slide.",
)
@click.option(
    "--num-processes",
    type=int,
    default=4,
    help="Number of processes to use when `--make-color-text` is enabled.",
)
def tosbu(
    *,
    results_dir: Path,
    output: Path,
    wsi_dir: Path,
    execution_id: str,
    study_id: str,
    make_color_text: bool = False,
    num_processes: int = 4,
) -> None:
    """Convert model outputs to Stony Brook format.

    Convert CSVs of patch predictions to .txt and .json formats for use with Stony
    Brook Biomedical Informatics viewers.

    RESULTS_DIR     Path to results directory (containing model-outputs dir).

    OUTPUT          Path to output directory in which to save files.
    """
    if not results_dir.exists():
        raise click.ClickException(f"results_dir does not exist: {results_dir}")
    if output.exists():
        raise click.ClickException("Output directory already exists.")
    if (
        not (results_dir / "model-outputs").exists()
        and (results_dir / "patches").exists()
    ):
        raise click.ClickException(
            "Model outputs have not been generated yet. Please run model inference."
        )
    if not (results_dir / "model-outputs").exists():
        raise click.ClickException(
            "Expected results_dir to contain a 'model-outputs' directory but it does"
            " not. Please provide the path to the directory that contains"
            " model-outputs, masks, and patches."
        )
    if not (results_dir / "run_metadata.json").exists():
        raise click.ClickException(
            f"Cannot find {results_dir / 'run_metadata.json'}. Was model inference run?"
        )
    if not wsi_dir.exists():
        raise click.ClickException("Whole slide image directory does not exist.")

    csvs = list((results_dir / "model-outputs").glob("*.csv"))
    if not csvs:
        raise click.ClickException("No CSVs found. Did you generate model outputs?")

    output.mkdir(exist_ok=False)

    with open(results_dir / "run_metadata.json") as f:
        run_metadata: dict = json.load(f)

    print("-" * 40)
    print("Run metadata")
    pprint.pprint(run_metadata, indent=2)
    print("-" * 40, "\n")

    # Get the output classes of the model. We will create separate directories for each
    # label name. But by default, we do not include labels that start with "no", like
    # "notils" and "notumor".
    class_names: list[str] = run_metadata["model_weights"]["class_names"]
    ignore_names = {"notils", "notumor"}
    class_names = [n for n in class_names if n not in ignore_names]

    print("-" * 40)
    print("Output class names:")
    for c in class_names:
        print(f"  - {c}")
    print("-" * 40)

    for ii, input_csv in enumerate(csvs):
        print("-" * 40)
        print(f"Converting outputs for slide {ii+1} of {len(csvs)}")
        slide_id = input_csv.stem
        wsi_file = wsi_dir / f"{slide_id}.svs"
        if not wsi_file.exists():
            click.secho(f"WSI file not found: {wsi_file}", bg="red")
            click.secho("Skipping...", bg="red")
            continue
        slide = WSI(wsi_file)

        slide_width, slide_height = slide.level_dimensions[0]

        for class_name in class_names:
            if len(class_names) == 1:
                # Do not use class-specific dirs if there is only one class.
                output_heatmap = output / "heatmap_json" / f"heatmap_{slide_id}.json"
                output_meta = output_heatmap.parent / f"meta_{slide_id}.json"
            else:
                output_heatmap = (
                    output / "heatmap_json" / class_name / f"heatmap_{slide_id}.json"
                )
            output_meta = output_heatmap.parent / f"meta_{slide_id}.json"
            output_heatmap.parent.mkdir(parents=True, exist_ok=True)

            click.echo("Writing JSON heatmap:")
            click.echo(f"  {output_heatmap}")
            click.echo("Writing JSON meta:")
            click.echo(f"  {output_meta}")
            write_heatmap_and_meta_json_lines(
                input=input_csv,
                output_heatmap=output_heatmap,
                output_meta=output_meta,
                slide_width=slide_width,
                slide_height=slide_height,
                execution_id=execution_id,
                study_id=study_id,
                case_id=slide_id,  # TODO: should case_id be different?
                subject_id=slide_id,
                class_name=class_name,
                run_metadata=run_metadata,
            )

            # Do not use class-specific dirs if there is only one class.
            if len(class_names) == 1:
                output_txt_prediction = (
                    output / "heatmap_txt" / f"prediction-{slide_id}"
                )
            else:
                output_txt_prediction = (
                    output / "heatmap_txt" / class_name / f"prediction-{slide_id}"
                )
            output_txt_prediction.parent.mkdir(parents=True, exist_ok=True)
            print("Writing TXT heatmap:")
            print(f"  {output_txt_prediction}")
            write_heatmap_txt(
                input=input_csv, output=output_txt_prediction, class_names=[class_name]
            )

        # Do this once per WSI. If the model has multi-class outputs, copy the color
        # file into each class-specific directory.
        if make_color_text:
            print("Writing color TXT file with color information for each patch")
            if len(class_names) == 1:
                output_color = output / "heatmap_txt" / f"color-{slide_id}"
                print(f"  {output_color}")
                write_color_txt(
                    input=input_csv,
                    output=output_color,
                    slide=slide,
                    num_processes=num_processes,
                )
            else:
                output_color = (
                    output / "heatmap_txt" / class_names[0] / f"color-{slide_id}"
                )
                print(f"  {output_color}")
                write_color_txt(
                    input=input_csv,
                    output=output_color,
                    slide=slide,
                    num_processes=num_processes,
                )
                # Copy this color file to all class-specific dirs.
                print("Copying color TXT file to class-specific TXT directories")
                for class_name in class_names[1:]:
                    cp_dst = output_color.parent.parent / class_name / output_color.name
                    print(f"  {cp_dst}")
                    shutil.copy(output_color, cp_dst)

    click.secho("Finished.", fg="green")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import wsinfer

project = "WSInfer"
copyright = "2023, Jakub Kaczmarzyk"
author = "Jakub Kaczmarzyk"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.intersphinx",  # for links
    "sphinx.ext.napoleon",  # for google style docstrings
    "sphinx.ext.viewcode",  # add links to code
    "autoapi.extension",  # to document the wsinfer api
    "sphinx_click",  # to document click command line
    "sphinx_copybutton",  # add copy button to top-right of code blocks
]

# Internationalization.
language = "en"

# AutoAPI options.
autoapi_type = "python"
autoapi_dirs = ["../wsinfer"]
autoapi_options = [
    "members",
    "undoc-members",
    # "private-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
    "imported-members",
]
autoapi_ignore = ["*cli*", "*__main__.py"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "torch": ("https://pytorch.org/docs/stable", None),
    "PIL": ("https://pillow.readthedocs.io/en/stable/", None),
    "openslide": ("https://openslide.org/api/python/", None),
}

# For editing the pages.
html_context = {
    "github_user": "SBU-BMI",
    "github_repo": "wsinfer",
    "github_version": "main",
    "doc_path": "docs",
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# Define the json_url for our version switcher.
json_url = "https://wsinfer.readthedocs.io/en/latest/_static/switcher.json"

# Copied (with love) from conf.py of pydata-sphinx-theme.
# Define the version we use for matching in the version switcher.
version_match = os.environ.get("READTHEDOCS_VERSION")
# If READTHEDOCS_VERSION doesn't exist, we're not on RTD
# If it is an integer, we're in a PR build and the version isn't correct.
if not version_match or version_match.isdigit():
    # For local development, infer the version to match from the package.
    release = wsinfer.__version__
    if "dev" in release or "rc" in release:
        version_match = "dev"
        # We want to keep the relative reference if we are in dev mode
        # but we want the whole url if we are effectively in a released version
        json_url = "_static/switcher.json"
    else:
        version_match = "v" + release

html_theme = "pydata_sphinx_theme"
# Many of these options are copied directly from pydata-sphinx-theme's conf.py.
html_theme_options = {
    "logo": {
        "text": "WSInfer",
        # "image_dark": "_static/logo-dark.svg",
        # "alt_text": "WSInfer",
    },
    "use_edit_page_button": True,
    "show_toc_level": 1,
    "navbar_align": "left",  # [left, content, right] For testing that the navbar items align properly
    "navbar_center": ["version-switcher", "navbar-nav"],
    "show_version_warning_banner": True,
    # "show_nav_level": 2,
    # "navbar_start": ["navbar-logo"],
    # "navbar_end": ["theme-switcher", "navbar-icon-links"],
    # "navbar_persistent": ["search-button"],
    # "primary_sidebar_end": ["custom-template.html", "sidebar-ethical-ads.html"],
    # "article_footer_items": ["test.html", "test.html"],
    # "content_footer_items": ["test.html", "test.html"],
    # "footer_start": ["test.html", "test.html"],
    # "secondary_sidebar_items": ["page-toc.html"],  # Remove the source buttons
    "switcher": {
        "json_url": json_url,
        "version_match": version_match,
    },
}
# html_logo = "_static/logo.svg"
# html_favicon = "_static/log



"""PyTorch image classification transform."""

from __future__ import annotations

from torchvision import transforms
from wsinfer_zoo.client import TransformConfigurationItem

# The subset of transforms known to the wsinfer config spec.
# This can be expanded in the future as needs arise.
_name_to_tv_cls = {
    "Resize": transforms.Resize,
    "ToTensor": transforms.ToTensor,
    "Normalize": transforms.Normalize,
}


def make_compose_from_transform_config(
    list_of_transforms: list[TransformConfigurationItem],
) -> transforms.Compose:
    """Create a torchvision Compose instance from configuration of transforms."""
    all_t: list = []
    for t in list_of_transforms:
        cls = _name_to_tv_cls[t.name]
        kwargs = t.arguments or {}
        all_t.append(cls(**kwargs))
    return transforms.Compose(all_t)



from __future__ import annotations

import logging
from typing import Literal

import click

from ..wsi import set_backend
from .convert_csv_to_geojson import togeojson
from .convert_csv_to_sbubmi import tosbu
from .infer import run
from .patch import patch

_logging_levels = ["debug", "info", "warning", "error", "critical"]


# We use invoke_without_command=True so that 'wsinfer' on its own can be used for
# inference on slides.
@click.group()
@click.option(
    "--backend",
    default=None,
    help="Backend for loading whole slide images.",
    type=click.Choice(["openslide", "tiffslide"]),
)
@click.option(
    "--log-level",
    default="info",
    type=click.Choice(_logging_levels),
    help="Set the loudness of logging.",
)
@click.version_option()
def cli(
    backend: Literal["openslide"] | Literal["tiffslide"] | None, log_level: str
) -> None:
    """Run patch-level classification inference on whole slide images."""

    # Configure logger.
    levels = {level: getattr(logging, level.upper()) for level in _logging_levels}
    level = levels[log_level]
    logging.basicConfig(level=level)

    if backend is not None:
        set_backend(backend)


cli.add_command(run)
cli.add_command(togeojson)
cli.add_command(tosbu)
cli.add_command(patch)



from __future__ import annotations

import dataclasses
import warnings
from typing import Callable

import torch
import wsinfer_zoo
from wsinfer_zoo.client import HFModelTorchScript
from wsinfer_zoo.client import Model


@dataclasses.dataclass
class LocalModelTorchScript(Model):
    ...


def get_registered_model(name: str) -> HFModelTorchScript:
    registry = wsinfer_zoo.client.load_registry()
    model = registry.get_model_by_name(name=name)
    return model.load_model_torchscript()


def get_pretrained_torch_module(
    model: HFModelTorchScript | LocalModelTorchScript,
) -> torch.nn.Module:
    """Get a PyTorch Module with weights loaded."""
    mod: torch.nn.Module = torch.jit.load(model.model_path, map_location="cpu")
    if not isinstance(mod, torch.nn.Module):
        raise TypeError(
            "expected the loaded object to be a subclass of torch.nn.Module but got"
            f" {type(mod)}."
        )
    return mod


def jit_compile(
    model: torch.nn.Module,
) -> torch.jit.ScriptModule | torch.nn.Module | Callable:
    """JIT-compile a model for inference.

    A torchscript model may be JIT compiled here as well.
    """
    noncompiled = model
    device = next(model.parameters()).device
    # Attempt to script. If it fails, return the original.
    test_input = torch.ones(1, 3, 224, 224).to(device)
    w = "Warning: could not JIT compile the model. Using non-compiled model instead."

    # PyTorch 2.x has torch.compile but it does not work when applied
    # to TorchScript models.
    if hasattr(torch, "compile") and not isinstance(model, torch.jit.ScriptModule):
        # Try to get the most optimized model.
        try:
            return torch.compile(model, fullgraph=True, mode="max-autotune")
        except Exception:
            pass
        try:
            return torch.compile(model, mode="max-autotune")
        except Exception:
            pass
        try:
            return torch.compile(model)
        except Exception:
            warnings.warn(w)
            return noncompiled
    # For pytorch 1.x, use torch.jit.script.
    else:
        try:
            mjit = torch.jit.script(model)
            with torch.no_grad():
                mjit(test_input)
        except Exception:
            warnings.warn(w)
            return noncompiled
        # Now that we have scripted the model, try to optimize it further. If that
        # fails, return the scripted model.
        try:
            mjit_frozen = torch.jit.freeze(mjit)
            mjit_opt = torch.jit.optimize_for_inference(mjit_frozen)
            with torch.no_grad():
                mjit_opt(test_input)
            return mjit_opt
        except Exception:
            return mjit  # type: ignore



from __future__ import annotations

import itertools
import logging
from typing import Sequence

import cv2 as cv
import numpy as np
import numpy.typing as npt
from shapely import MultiPolygon
from shapely import Point
from shapely import Polygon
from shapely import STRtree

logger = logging.getLogger(__name__)


def get_multipolygon_from_binary_arr(
    arr: npt.NDArray[np.int_], scale: tuple[float, float] | None = None
) -> tuple[MultiPolygon, Sequence[npt.NDArray[np.int_]], npt.NDArray[np.int_]]:
    """Create a Shapely Polygon from a binary array.

    Parameters
    ----------
    arr : array
        Binary array where non-zero values indicate presence of tissue.
    scale : tuple of two floats, optional
        If specified, this is the factor by which coordinates are multiplied to recover
        the coordinates at the base resolution of the whole slide image.

    Returns
    -------
    polygon
        A shapely `MultiPolygon` object representing tissue regions.
    contours
        A sequence of arrays representing unscaled contours of tissue.
    hierarchy
        An array of the hierarchy of contours.
    """
    # Find contours and hierarchy
    contours: Sequence[npt.NDArray]
    hierarchy: npt.NDArray[np.int_]
    contours, hierarchy = cv.findContours(arr, cv.RETR_CCOMP, cv.CHAIN_APPROX_SIMPLE)
    hierarchy = hierarchy.squeeze(0)

    logger.info(f"Detected {len(contours)} contours")
    contours_unscaled = contours
    if scale is not None:
        logger.info(
            "Scaling contours to slide-level (level=0) coordinate space by multiplying"
            f" by {scale}"
        )
        # Reshape to broadcast with contour coordinates.
        scale_arr: npt.NDArray[np.float_] = np.array(scale).reshape(1, 1, 2)
        contours = tuple(c * scale_arr for c in contours_unscaled)
        del scale_arr

    # From https://stackoverflow.com/a/75510437/5666087
    def merge_polygons(polygon: MultiPolygon, idx: int, add: bool) -> MultiPolygon:
        """
        polygon: Main polygon to which a new polygon is added
        idx: Index of contour
        add: If this contour should be added (True) or subtracted (False)
        """

        # Get contour from global list of contours
        contour = np.squeeze(contours[idx])

        # cv2.findContours() sometimes returns a single point -> skip this case
        if len(contour) > 2:
            # Convert contour to shapely polygon
            new_poly = Polygon(contour)

            # Not all polygons are shapely-valid (self intersection, etc.)
            if not new_poly.is_valid:
                # Convert invalid polygon to valid
                new_poly = new_poly.buffer(0)

            # Merge new polygon with the main one
            if add:
                polygon = polygon.union(new_poly)
            else:
                polygon = polygon.difference(new_poly)

        # Check if current polygon has a child
        child_idx = hierarchy[idx][2]
        if child_idx >= 0:
            # Call this function recursively, negate `add` parameter
            polygon = merge_polygons(polygon, child_idx, not add)

        # Check if there is some next polygon at the same hierarchy level
        next_idx = hierarchy[idx][0]
        if next_idx >= 0:
            # Call this function recursively
            polygon = merge_polygons(polygon, next_idx, add)

        return polygon

    # Call the function with an initial empty polygon and start from contour 0
    polygon = merge_polygons(MultiPolygon(), 0, True)

    # Add back the axis in hierarchy because we squeezed it before.
    return polygon, contours_unscaled, hierarchy[np.newaxis]


def get_nonoverlapping_patch_coordinates_within_polygon(
    slide_width: int,
    slide_height: int,
    patch_size: int,
    half_patch_size: int,
    polygon: Polygon,
) -> npt.NDArray[np.int_]:
    """Get coordinates of patches within a polygon.

    Parameters
    ----------
    slide_width : int
        The width of the slide in pixels at base resolution.
    slide_height : int
        The height of the slide in pixels at base resolution.
    patch_size : int
        The size of a patch in pixels.
    half_patch_size : int
        Half of the length of a patch in pixels.
    polygon : Polygon
        A shapely Polygon representing the presence of tissue.

    Returns
    -------
    coordinates
        Array with shape (N, 2), where N is the number of tiles. Each row in this array
        contains the coordinates of the top-left of a tile: (minx, miny).
    """

    # Make an array of Nx2, where each row is (x, y) centroid of the patch.
    tile_centroids_arr: npt.NDArray[np.int_] = np.array(
        list(
            itertools.product(
                range(0 + half_patch_size, slide_width, patch_size),
                range(0 + half_patch_size, slide_height, patch_size),
            )
        )
    )

    tile_centroids_poly = [Point(c) for c in tile_centroids_arr]

    # Query which centroids are inside the polygon.
    tree = STRtree(tile_centroids_poly)
    centroid_indexes_in_polygon: npt.NDArray[np.int_] = tree.query(
        polygon, predicate="contains"
    )

    # Sort so x and y are in ascending order (and y changes most rapidly).
    centroid_indexes_in_polygon.sort()
    tile_centroids_in_polygon = tile_centroids_arr[centroid_indexes_in_polygon]

    # Transform the centroids to the upper-left point (x, y).
    tile_minx_miny_in_polygon = tile_centroids_in_polygon - half_patch_size

    return tile_minx_miny_in_polygon



from __future__ import annotations

import json
import os
import platform
import sys
import time
from pathlib import Path

import geojson as geojsonlib
import h5py
import numpy as np
import pandas as pd
import pytest
import tifffile
import torch
from click.testing import CliRunner

from wsinfer.cli.cli import cli
from wsinfer.cli.infer import _get_info_for_save
from wsinfer.modellib.models import get_pretrained_torch_module
from wsinfer.modellib.models import get_registered_model
from wsinfer.modellib.run_inference import jit_compile


@pytest.fixture
def tiff_image(tmp_path: Path) -> Path:
    x = np.empty((4096, 4096, 3), dtype="uint8")
    x[...] = [160, 32, 240]  # rgb for purple
    path = Path(tmp_path / "images" / "purple.tif")
    path.parent.mkdir(exist_ok=True)

    tifffile.imwrite(
        path,
        data=x,
        compression="zlib",
        tile=(256, 256),
        # 0.25 micrometers per pixel.
        resolution=(40_000, 40_000),
        resolutionunit=tifffile.RESUNIT.CENTIMETER,
    )

    return path


# The reference data for this test was made using a patched version of wsinfer 0.3.6.
# The patches fixed an issue when calculating strides and added padding to images.
# Large-image (which was the backend in 0.3.6) did not pad images and would return
# tiles that were not fully the requested width and height.
@pytest.mark.parametrize(
    "model",
    [
        "breast-tumor-resnet34.tcga-brca",
        "breast-tumor-inception_v4.tcga-brca",
        "breast-tumor-vgg16mod.tcga-brca",
        "lung-tumor-resnet34.tcga-luad",
        "pancancer-lymphocytes-inceptionv4.tcga",
        "pancreas-tumor-preactresnet34.tcga-paad",
        "prostate-tumor-resnet34.tcga-prad",
    ],
)
@pytest.mark.parametrize("speedup", [False, True])
@pytest.mark.parametrize("backend", ["openslide", "tiffslide"])
def test_cli_run_with_registered_models(
    model: str,
    speedup: bool,
    backend: str,
    tiff_image: Path,
    tmp_path: Path,
):
    """A regression test of the command 'wsinfer run'."""

    reference_csv = Path(__file__).parent / "reference" / model / "purple.csv"
    if not reference_csv.exists():
        raise FileNotFoundError(f"reference CSV not found: {reference_csv}")

    runner = CliRunner()
    results_dir = tmp_path / "inference"
    result = runner.invoke(
        cli,
        [
            "--backend",
            backend,
            "run",
            "--wsi-dir",
            str(tiff_image.parent),
            "--results-dir",
            str(results_dir),
            "--model",
            model,
            "--speedup" if speedup else "--no-speedup",
        ],
    )
    assert result.exit_code == 0
    assert (results_dir / "model-outputs").exists()
    df = pd.read_csv(results_dir / "model-outputs" / "purple.csv")
    df_ref = pd.read_csv(reference_csv)

    assert set(df.columns) == set(df_ref.columns)
    assert df.shape == df_ref.shape
    assert np.array_equal(df["minx"], df_ref["minx"])
    assert np.array_equal(df["miny"], df_ref["miny"])
    assert np.array_equal(df["width"], df_ref["width"])
    assert np.array_equal(df["height"], df_ref["height"])

    prob_cols = df_ref.filter(like="prob_").columns.tolist()
    for prob_col in prob_cols:
        assert np.allclose(
            df[prob_col], df_ref[prob_col], atol=1e-07
        ), f"Column {prob_col} not allclose at atol=1e-07"

    # Test that metadata path exists.
    metadata_paths = list(results_dir.glob("run_metadata_*.json"))
    assert len(metadata_paths) == 1
    metadata_path = metadata_paths[0]
    assert metadata_path.exists()
    with open(metadata_path) as f:
        meta = json.load(f)
    assert set(meta.keys()) == {"model", "runtime", "timestamp"}
    assert "config" in meta["model"]
    assert "huggingface_location" in meta["model"]
    assert model in meta["model"]["huggingface_location"]["repo_id"]
    assert meta["runtime"]["python_executable"] == sys.executable
    assert meta["runtime"]["python_version"] == platform.python_version()
    assert meta["timestamp"]
    del metadata_path, meta

    # Test conversion to geojson.
    geojson_dir = results_dir / "geojson"
    result = runner.invoke(cli, ["togeojson", str(results_dir), str(geojson_dir)])
    assert result.exit_code == 0
    with open(geojson_dir / "purple.json") as f:
        d: geojsonlib.GeoJSON = geojsonlib.load(f)
    assert d.is_valid, "geojson not valid!"
    assert len(d["features"]) == len(df_ref)

    for geojson_row in d["features"]:
        assert geojson_row["type"] == "Feature"
        assert geojson_row["id"] == "PathTileObject"
        assert geojson_row["geometry"]["type"] == "Polygon"

    res = []
    for i, _ in enumerate(prob_cols):
        res.append(
            np.array(
                [dd["properties"]["measurements"][i]["value"] for dd in d["features"]]
            )
        )
    geojson_probs = np.stack(res, axis=0)
    del res
    assert np.allclose(df[prob_cols].T, geojson_probs)

    # Check the coordinate values.
    for df_row, geojson_row in zip(df.itertuples(), d["features"]):
        maxx = df_row.minx + df_row.width
        maxy = df_row.miny + df_row.height
        df_coords = [
            [maxx, df_row.miny],
            [maxx, maxy],
            [df_row.minx, maxy],
            [df_row.minx, df_row.miny],
            [maxx, df_row.miny],
        ]
        assert [df_coords] == geojson_row["geometry"]["coordinates"]


def test_cli_run_with_local_model(tmp_path: Path, tiff_image: Path):
    model = "breast-tumor-resnet34.tcga-brca"
    reference_csv = Path(__file__).parent / "reference" / model / "purple.csv"
    if not reference_csv.exists():
        raise FileNotFoundError(f"reference CSV not found: {reference_csv}")
    w = get_registered_model(model)

    config = {
        "spec_version": "1.0",
        "architecture": "resnet34",
        "num_classes": 2,
        "class_names": ["notumor", "tumor"],
        "patch_size_pixels": 350,
        "spacing_um_px": 0.25,
        "transform": [
            {"name": "Resize", "arguments": {"size": 224}},
            {"name": "ToTensor"},
            {
                "name": "Normalize",
                "arguments": {
                    "mean": [0.7238, 0.5716, 0.6779],
                    "std": [0.112, 0.1459, 0.1089],
                },
            },
        ],
    }

    config_path = tmp_path / "config.json"
    with open(config_path, "w") as f:
        json.dump(config, f)

    runner = CliRunner()
    results_dir = tmp_path / "inference"
    result = runner.invoke(
        cli,
        [
            "--backend",
            "openslide",
            "run",
            "--wsi-dir",
            str(tiff_image.parent),
            "--results-dir",
            str(results_dir),
            "--model-path",
            w.model_path,
            "--config",
            str(config_path),
        ],
    )
    assert result.exit_code == 0
    assert (results_dir / "model-outputs").exists()
    df = pd.read_csv(results_dir / "model-outputs" / "purple.csv")
    df_ref = pd.read_csv(reference_csv)

    assert set(df.columns) == set(df_ref.columns)
    assert df.shape == df_ref.shape
    assert np.array_equal(df["minx"], df_ref["minx"])
    assert np.array_equal(df["miny"], df_ref["miny"])
    assert np.array_equal(df["width"], df_ref["width"])
    assert np.array_equal(df["height"], df_ref["height"])

    prob_cols = df_ref.filter(like="prob_").columns.tolist()
    for prob_col in prob_cols:
        assert np.allclose(
            df[prob_col], df_ref[prob_col], atol=1e-07
        ), f"Column {prob_col} not allclose at atol=1e-07"


def test_cli_run_no_model_or_config(tmp_path: Path):
    """Test that --model or (--config and --model-path) is required."""
    wsi_dir = tmp_path / "slides"
    wsi_dir.mkdir()

    runner = CliRunner()
    args = [
        "run",
        "--wsi-dir",
        str(wsi_dir),
        "--results-dir",
        str(tmp_path / "results"),
    ]
    # No model, weights, or config.
    result = runner.invoke(cli, args)
    assert result.exit_code != 0
    assert "one of --model or (--config and --model-path) is required" in result.output


def test_cli_run_model_and_config(tmp_path: Path):
    """Test that (model and weights) or config is required."""
    wsi_dir = tmp_path / "slides"
    wsi_dir.mkdir()

    fake_config = tmp_path / "foobar.json"
    fake_config.touch()
    fake_model_path = tmp_path / "foobar.pt"
    fake_model_path.touch()

    runner = CliRunner()
    args = [
        "run",
        "--wsi-dir",
        str(wsi_dir),
        "--results-dir",
        str(tmp_path / "results"),
        "--model",
        "colorectal-tiatoolbox-resnet50.kather100k",
        "--model-path",
        str(fake_model_path),
        "--config",
        str(fake_config),
    ]
    # No model, weights, or config.
    result = runner.invoke(cli, args)
    assert result.exit_code != 0
    assert (
        "--config and --model-path are mutually exclusive with --model" in result.output
    )


@pytest.mark.xfail
def test_convert_to_sbu():
    # TODO: create a synthetic output and then convert it. Check that it is valid.
    assert False


@pytest.mark.parametrize(
    ["patch_size", "patch_spacing"],
    [(256, 0.25), (256, 0.50), (350, 0.25), (100, 0.3), (100, 0.5)],
)
@pytest.mark.parametrize("backend", ["openslide", "tiffslide"])
def test_patch_cli(
    patch_size: int,
    patch_spacing: float,
    backend: str,
    tmp_path: Path,
    tiff_image: Path,
):
    """Test of 'wsinfer patch'."""
    orig_slide_size = 4096
    orig_slide_spacing = 0.25

    runner = CliRunner()
    savedir = tmp_path / "savedir"
    result = runner.invoke(
        cli,
        [
            "--backend",
            backend,
            "patch",
            "--wsi-dir",
            str(tiff_image.parent),
            "--results-dir",
            str(savedir),
            "--patch-size-px",
            str(patch_size),
            "--patch-spacing-um-px",
            str(patch_spacing),
        ],
    )
    assert result.exit_code == 0
    stem = tiff_image.stem
    assert (savedir / "masks" / f"{stem}.jpg").exists()
    assert (savedir / "patches" / f"{stem}.h5").exists()

    expected_patch_size = round(patch_size * patch_spacing / orig_slide_spacing)
    sqrt_expected_num_patches = round(orig_slide_size / expected_patch_size)
    expected_num_patches = sqrt_expected_num_patches**2

    expected_coords = []
    for x in range(0, orig_slide_size, expected_patch_size):
        for y in range(0, orig_slide_size, expected_patch_size):
            # Patch is kept if centroid is inside.
            if (
                x + expected_patch_size // 2 <= orig_slide_size
                and y + expected_patch_size // 2 <= orig_slide_size
            ):
                expected_coords.append([x, y])
    assert len(expected_coords) == expected_num_patches
    with h5py.File(savedir / "patches" / f"{stem}.h5") as f:
        assert f["/coords"].attrs["patch_size"] == expected_patch_size
        coords = f["/coords"][()]
    assert coords.shape == (expected_num_patches, 2)
    assert np.array_equal(expected_coords, coords)


# FIXME: parametrize this test across our models.
def test_jit_compile():
    w = get_registered_model("breast-tumor-resnet34.tcga-brca")
    model = get_pretrained_torch_module(w)

    x = torch.ones(20, 3, 224, 224, dtype=torch.float32)
    model.eval()
    NUM_SAMPLES = 1
    with torch.no_grad():
        t0 = time.perf_counter()
        for _ in range(NUM_SAMPLES):
            out_nojit = model(x).detach().cpu()
        time_nojit = time.perf_counter() - t0
    model_nojit = model
    model = jit_compile(model)  # type: ignore
    if model is model_nojit:
        pytest.skip("Failed to compile model (would use original model)")
    with torch.no_grad():
        model(x).detach().cpu()  # run it once to compile
        t0 = time.perf_counter()
        for _ in range(NUM_SAMPLES):
            out_jit = model(x).detach().cpu()
        time_yesjit = time.perf_counter() - t0

    assert torch.allclose(out_nojit, out_jit)
    if time_nojit < time_yesjit:
        pytest.skip(
            "JIT-compiled model was SLOWER than original: "
            f"jit={time_yesjit:0.3f} vs nojit={time_nojit:0.3f}"
        )


def test_issue_89():
    """Do not fail if 'git' is not installed."""
    model_obj = get_registered_model("breast-tumor-resnet34.tcga-brca")
    d = _get_info_for_save(model_obj)
    assert d
    assert "git" in d["runtime"]
    assert d["runtime"]["git"]
    assert d["runtime"]["git"]["git_remote"]
    assert d["runtime"]["git"]["git_branch"]

    # Test that _get_info_for_save does not fail if git is not found.
    orig_path = os.environ["PATH"]
    try:
        os.environ["PATH"] = ""
        d = _get_info_for_save(model_obj)
        assert d
        assert "git" in d["runtime"]
        assert d["runtime"]["git"] is None
    finally:
        os.environ["PATH"] = orig_path  # reset path


def test_issue_94(tmp_path: Path, tiff_image: Path):
    """Gracefully handle unreadable slides."""

    # We have a valid tiff in 'tiff_image.parent'. We put in an unreadable file too.
    badpath = tiff_image.parent / "bad.svs"
    badpath.touch()

    runner = CliRunner()
    results_dir = tmp_path / "inference"
    result = runner.invoke(
        cli,
        [
            "run",
            "--wsi-dir",
            str(tiff_image.parent),
            "--results-dir",
            str(results_dir),
            "--model",
            "breast-tumor-resnet34.tcga-brca",
        ],
    )
    # Important part is that we run through all of the files, despite the unreadble
    # file.
    assert result.exit_code == 0
    assert results_dir.joinpath("model-outputs").joinpath("purple.csv").exists()
    assert not results_dir.joinpath("model-outputs").joinpath("bad.csv").exists()


def test_issue_97(tmp_path: Path, tiff_image: Path):
    """Write a run_metadata file per run."""

    runner = CliRunner()
    results_dir = tmp_path / "inference"
    result = runner.invoke(
        cli,
        [
            "run",
            "--wsi-dir",
            str(tiff_image.parent),
            "--results-dir",
            str(results_dir),
            "--model",
            "breast-tumor-resnet34.tcga-brca",
        ],
    )
    assert result.exit_code == 0
    metas = list(results_dir.glob("run_metadata_*.json"))
    assert len(metas) == 1

    time.sleep(2)  # make sure some time has passed so the timestamp is different

    # Run again...
    result = runner.invoke(
        cli,
        [
            "run",
            "--wsi-dir",
            str(tiff_image.parent),
            "--results-dir",
            str(results_dir),
            "--model",
            "breast-tumor-resnet34.tcga-brca",
        ],
    )
    assert result.exit_code == 0
    metas = list(results_dir.glob("run_metadata_*.json"))
    assert len(metas) == 2


def test_issue_125(tmp_path: Path):
    """Test that path in model config can be saved when a pathlib.Path object."""

    w = get_registered_model("breast-tumor-resnet34.tcga-brca")
    w.model_path = Path(w.model_path)  # type: ignore
    info = _get_info_for_save(w)
    with open(tmp_path / "foo.json", "w") as f:
        json.dump(info, f)



from __future__ import annotations

from pathlib import Path

import click

from ..patchlib import segment_and_patch_directory_of_slides


@click.command()
@click.option(
    "-i",
    "--wsi-dir",
    type=click.Path(exists=True, file_okay=False, path_type=Path, resolve_path=True),
    required=True,
    help="Directory containing whole slide images. This directory can *only* contain"
    " whole slide images.",
)
@click.option(
    "-o",
    "--results-dir",
    type=click.Path(file_okay=False, path_type=Path, resolve_path=True),
    required=True,
    help="Directory to store patch results. If directory exists, will skip"
    " whole slides for which outputs exist.",
)
@click.option("--patch-size-px", required=True, type=int, help="Patch size in pixels.")
@click.option(
    "--patch-spacing-um-px",
    required=True,
    type=float,
    help="Physical spacing of the patch in micrometers per pixel.",
)
@click.option(
    "--thumbsize",
    default=(2048, 2048),
    type=(int, int),
    help="The size of the slide thumbnail (in pixels) used for tissue segmentation."
    " The aspect ratio is preserved, and the longest side will have length"
    " max(thumbsize).",
)
@click.option(
    "--median-filter-size",
    default=7,
    type=click.IntRange(min=3),
    help="The kernel size for median filtering. Must be greater than 1 and odd.",
)
@click.option(
    "--binary-threshold",
    default=7,
    type=click.IntRange(min=1),
    help="The threshold for image binarization.",
)
@click.option(
    "--closing-kernel-size",
    default=6,
    type=click.IntRange(min=1),
    help="The kernel size for binary closing (morphological operation).",
)
@click.option(
    "--min-object-size-um2",
    default=200**2,
    type=click.FloatRange(min=0),
    help="The minimum size of an object to keep during tissue detection. If a"
    " contiguous object is smaller than this area, it replaced with background."
    " The default is 200um x 200um. The units of this argument are microns squared.",
)
@click.option(
    "--min-hole-size-um2",
    default=190**2,
    type=click.FloatRange(min=0),
    help="The minimum size of a hole to keep as a hole. If a hole is smaller than this"
    " area, it is filled with foreground. The default is 190um x 190um. The units of"
    " this argument are microns squared.",
)
def patch(
    wsi_dir: str,
    results_dir: str,
    patch_size_px: int,
    patch_spacing_um_px: float,
    thumbsize: tuple[int, int],
    median_filter_size: int,
    binary_threshold: int,
    closing_kernel_size: int,
    min_object_size_um2: float,
    min_hole_size_um2: float,
) -> None:
    """Patch a directory of whole slide iamges."""
    segment_and_patch_directory_of_slides(
        wsi_dir=wsi_dir,
        save_dir=results_dir,
        patch_size_px=patch_size_px,
        patch_spacing_um_px=patch_spacing_um_px,
        thumbsize=thumbsize,
        median_filter_size=median_filter_size,
        binary_threshold=binary_threshold,
        closing_kernel_size=closing_kernel_size,
        min_object_size_um2=min_object_size_um2,
        min_hole_size_um2=min_hole_size_um2,
    )



from __future__ import annotations

import logging
from pathlib import Path
from typing import Sequence

import cv2 as cv
import h5py
import numpy as np
import numpy.typing as npt
from PIL import Image

from ..wsi import WSI
from ..wsi import _validate_wsi_directory
from ..wsi import get_avg_mpp
from .patch import get_multipolygon_from_binary_arr
from .patch import get_nonoverlapping_patch_coordinates_within_polygon
from .segment import segment_tissue

logger = logging.getLogger(__name__)

MASKS_DIR = "masks"
PATCHES_DIR = "patches"


def segment_and_patch_one_slide(
    slide_path: str | Path,
    save_dir: str | Path,
    patch_size_px: int,
    patch_spacing_um_px: float,
    thumbsize: tuple[int, int] = (2048, 2048),
    median_filter_size: int = 7,
    binary_threshold: int = 7,
    closing_kernel_size: int = 6,
    min_object_size_um2: float = 200**2,
    min_hole_size_um2: float = 190**2,
) -> None:
    """Get non-overlapping patch coordinates in tissue regions of a whole slide image.

    Patch coordinates are saved to an HDF5 file in `{save_dir}/patches/`, and a tissue
    detection image is saved to `{save_dir}/masks/` for quality control.

    In general, this function takes the following steps:

    1. Get a low-resolution thumbnail of the image.
    2. Binarize the image to identify tissue regions.
    3. Process this binary image to remove artifacts.
    4. Create a regular grid of non-overlapping patches of specified size.
    5. Keep patches whose centroids are in tissue regions.

    Parameters
    ----------
    slide_path : str or Path
        The path to the whole slide image file.
    save_dir : str or Path
        The directory in which to save patching results.
    patch_size_px : int
        The length of one side of a square patch in pixels.
    patch_spacing_um_px : float
        The physical spacing of patches in micrometers per pixels. This value multiplied
        by patch_size_px gives the physical length of a patch in micrometers.
    thumbsize : tuple of two integers
        The size of the thumbnail to use for tissue detection. This specifies the
        largest possible bounding box of the thumbnail, and a thumbnail is taken to fit
        this space while maintaining the original aspect ratio of the whole slide image.
        Larger thumbnails will take longer to process but will result in better tissue
        masks.
    median_filter_size : int
        The size of the kernel for median filtering. This value must be odd and greater
        than one. This is in units of pixels in the thumbnail.
    binary_threshold: int
        The value at which the image in binarized. A higher value will keep less tissue.
    closing_kernel_size : int
        The size of the kernel for a morphological closing operation. This is in units
        of pixels in the thumbnail.
    min_object_size_um2 : float
        The minimum area of an object to keep, in units of micrometers squared. Any
        disconnected objects smaller than this area will be removed.
    min_hole_size_um2 : float
        The minimum size of a hole to keep, in units of micrometers squared. Any hole
        smaller than this area will be filled and be considered tissue.

    Returns
    -------
    None
    """

    save_dir = Path(save_dir).resolve()
    slide_path = Path(slide_path).resolve()
    slide_prefix = slide_path.stem

    logger.info(f"Segmenting and patching slide {slide_path}")
    logger.info(f"Using prefix as slide ID: {slide_prefix}")

    patch_path = save_dir / PATCHES_DIR / f"{slide_prefix}.h5"
    mask_path = save_dir / MASKS_DIR / f"{slide_prefix}.jpg"

    # End early if outputs exist.
    if patch_path.exists() and mask_path.exists():
        logger.info("Patch output and mask output files already exist")
        logger.info(f"patch_path={patch_path}")
        logger.info(f"mask_path={mask_path}")
        return None

    slide = WSI(slide_path)
    mpp = get_avg_mpp(slide_path)
    logger.info(f"Slide has WxH {slide.dimensions} and MPP={mpp}")

    logger.info(
        f"Requested patch size of {patch_size_px} px at {patch_spacing_um_px} um/px"
    )
    logger.info(
        f"Scaling patch size by {patch_spacing_um_px / mpp} for patch coordinates at"
        f" level 0 (MPP={mpp}) | patch_spacing_um_px / mpp"
        f" ({patch_spacing_um_px} / {mpp})"
    )
    patch_size = round(patch_size_px * patch_spacing_um_px / mpp)
    logger.info(f"Final patch size is {patch_size}")

    # Segment tissue into a binary mask.
    if len(thumbsize) != 2:
        raise ValueError(f"Length of 'thumbsize' must be 2 but got {len(thumbsize)}")
    thumb: Image.Image = slide.get_thumbnail(thumbsize)
    # TODO: allow the min hole size and min object size to be set in physical units.

    # thumb has ~12 MPP.
    thumb_mpp = (mpp * (np.array(slide.dimensions) / thumb.size)).mean()
    logger.info(f"Thumbnail has WxH {thumb.size} and MPP={thumb_mpp}")
    thumb_mpp_squared: float = thumb_mpp**2

    # (pixels2 / micron2) * micron2 = pixels2
    min_object_size_px: int = round(min_object_size_um2 / thumb_mpp_squared)
    min_hole_size_px: int = round(min_hole_size_um2 / thumb_mpp_squared)

    logger.info(
        f"Transformed minimum object size to {min_object_size_px} pixel area in"
        " thumbnail"
    )
    logger.info(
        f"Transformed minimum hole size to {min_hole_size_px} pixel area in thumbnail"
    )

    arr = segment_tissue(
        np.asarray(thumb),
        median_filter_size=median_filter_size,
        binary_threshold=binary_threshold,
        closing_kernel_size=closing_kernel_size,
        min_object_size_px=min_object_size_px,
        min_hole_size_px=min_hole_size_px,
    )
    if not np.issubdtype(arr.dtype, np.bool_):
        raise TypeError(
            f"expected the segmentation array to be boolean dtype but got {arr.dtype}"
        )

    # Create a polygon of the binary tissue mask.
    scale: tuple[float, float] = (
        slide.dimensions[0] / thumb.size[0],
        slide.dimensions[1] / thumb.size[1],
    )
    polygon, contours, hierarchy = get_multipolygon_from_binary_arr(
        arr.astype("uint8") * 255, scale=scale
    )

    # Get the coordinates of patches inside the tissue polygon.
    slide_width, slide_height = slide.dimensions
    half_patch_size = round(patch_size / 2)

    # Nx4 --> N x (minx, miny, width, height)
    coords = get_nonoverlapping_patch_coordinates_within_polygon(
        slide_width=slide_width,
        slide_height=slide_height,
        patch_size=patch_size,
        half_patch_size=half_patch_size,
        polygon=polygon,
    )
    logger.info(f"Found {len(coords)} patches within tissue")

    # Save coordinates to HDF5.
    patch_path.parent.mkdir(exist_ok=True, parents=True)
    save_hdf5(
        path=patch_path,
        coords=coords,
        patch_size=patch_size,
        patch_spacing_um_px=patch_spacing_um_px,
        compression="gzip",
    )

    # Save thumbnail with drawn contours.
    logger.info(f"Writing tissue thumbnail with contours to disk: {mask_path}")
    mask_path.parent.mkdir(exist_ok=True, parents=True)
    img = draw_contours_on_thumbnail(thumb, contours=contours, hierarchy=hierarchy)
    img.thumbnail((1024, 1024), resample=Image.Resampling.LANCZOS)
    img.save(mask_path)

    return None


def save_hdf5(
    path: str | Path,
    coords: npt.NDArray[np.int_],
    patch_size: int,
    patch_spacing_um_px: float,
    compression: str | None = "gzip",
) -> None:
    """Write patch coordinates to HDF5 file.

    This is designed to be interoperable with HDF5 files created by CLAM.

    Parameters
    ----------
    path : str or Path
        Path to save the HDF5 file.
    coords : array
        Nx2 array of coordinates, where N is the number of patches. Each row of the
        array must be minx and miny to specify the top-left of the patch.
    patch_size : int
        The size of patches in pixels at level 0 of the slide (base resolution).
    patch_spacing_um_px : float
        The physical spacing of the patch in micrometers per pixel.
    compression: str, optional
        Compression to use for storing coordinates. Default is "gzip".

    Returns
    -------
    None
    """
    logger.info(f"Writing coordinates to disk: {path}")
    logger.info(f"Coordinates have shape {coords.shape}")

    if coords.ndim != 2:
        raise ValueError(f"coords must have 2 dimensions but got {coords.ndim}")
    if coords.shape[1] != 2:
        raise ValueError(f"length of second axis must be 2 but got {coords.shape[1]}")

    with h5py.File(path, "w") as f:
        dset = f.create_dataset("/coords", data=coords, compression=compression)
        dset.attrs["patch_size"] = patch_size
        dset.attrs["patch_level"] = 0
        dset.attrs["patch_spacing_um_px"] = patch_spacing_um_px


def draw_contours_on_thumbnail(
    thumb: Image.Image,
    contours: Sequence[npt.NDArray[np.int_]],
    hierarchy: npt.NDArray[np.int_],
) -> Image.Image:
    """Draw contours onto an image.

    Parameters
    ----------
    thumb : Image.Image
        The thumbnail of the whole slide of the same size as the binary image used
        during contour detection.
    contours : sequence of arrays
        The contours result of `cv.findContours`.
    hierarchy : array
        The hierarchy result of `cv.findContours`.

    Returns
    -------
    Image.Image
        An image with contours burned in.
    """
    assert hierarchy.ndim == 3
    assert hierarchy.shape[0] == 1
    assert hierarchy.shape[2] == 4
    assert len(contours) == hierarchy.shape[1]

    # We assume the hierarchy was made with RETR_CCOMP.
    contour_is_external = (hierarchy[0, :, 3] < 0).tolist()
    external = [c for c, external in zip(contours, contour_is_external) if external]
    hole = [c for c, external in zip(contours, contour_is_external) if not external]

    img = np.array(thumb)
    # Colors are BGR.
    cv.drawContours(img, external, -1, (0, 255, 255), 7)
    cv.drawContours(img, hole, -1, (255, 255, 0), 7)

    return Image.fromarray(img).convert("RGB")


def segment_and_patch_directory_of_slides(
    wsi_dir: str | Path,
    save_dir: str | Path,
    patch_size_px: int,
    patch_spacing_um_px: float,
    thumbsize: tuple[int, int] = (2048, 2048),
    median_filter_size: int = 7,
    binary_threshold: int = 7,
    closing_kernel_size: int = 6,
    min_object_size_um2: float = 200**2,
    min_hole_size_um2: float = 190**2,
) -> None:
    """Get non-overlapping patch coordinates in tissue regions for a directory of whole
    slide images.

    Patch coordinates are saved to HDF5 files in `{save_dir}/patches/`, and tissue
    detection images are saved to `{save_dir}/masks/` for quality control.

    In general, this function takes the following steps for each whole slide image:

    1. Get a low-resolution thumbnail of the image.
    2. Binarize the image to identify tissue regions.
    3. Process this binary image to remove artifacts.
    4. Create a regular grid of non-overlapping patches of specified size.
    5. Keep patches whose centroids are in tissue regions.

    Parameters
    ----------
    wsi_dir : str or Path
        The directory of whole slide images. This must only contain whole slide images.
    save_dir : str or Path
        The directory in which to save patching results.
    patch_size_px : int
        The length of one side of a square patch in pixels.
    patch_spacing_um_px : float
        The physical spacing of patches in micrometers per pixels. This value multiplied
        by patch_size_px gives the physical length of a patch in micrometers.
    thumbsize : tuple of two integers
        The size of the thumbnail to use for tissue detection. This specifies the
        largest possible bounding box of the thumbnail, and a thumbnail is taken to fit
        this space while maintaining the original aspect ratio of the whole slide image.
        Larger thumbnails will take longer to process but will result in better tissue
        masks.
    median_filter_size : int
        The size of the kernel for median filtering. This value must be odd and greater
        than one. This is in units of pixels in the thumbnail.
    binary_threshold: int
        The value at which the image in binarized. A higher value will keep less tissue.
    closing_kernel_size : int
        The size of the kernel for a morphological closing operation. This is in units
        of pixels in the thumbnail.
    min_object_size_um2 : float
        The minimum area of an object to keep, in units of micrometers squared. Any
        disconnected objects smaller than this area will be removed.
    min_hole_size_um2 : float
        The minimum size of a hole to keep, in units of micrometers squared. Any hole
        smaller than this area will be filled and be considered tissue.

    Returns
    -------
    None
    """

    wsi_dir = Path(wsi_dir)

    _validate_wsi_directory(wsi_dir)

    slide_paths = sorted(wsi_dir.glob("*"))

    # NOTE: we could use multi-processing here but then the logs would get
    # discombobulated.
    for i, slide_path in enumerate(slide_paths):
        print(f"Slide {i+1} of {len(slide_paths)} ({(i+1)/len(slide_paths):.2%})")
        try:
            segment_and_patch_one_slide(
                slide_path=slide_path,
                save_dir=save_dir,
                patch_size_px=patch_size_px,
                patch_spacing_um_px=patch_spacing_um_px,
                thumbsize=thumbsize,
                median_filter_size=median_filter_size,
                binary_threshold=binary_threshold,
                closing_kernel_size=closing_kernel_size,
                min_object_size_um2=min_object_size_um2,
                min_hole_size_um2=min_hole_size_um2,
            )
        except Exception as e:
            logger.error(f"Failed to segment and patch slide\n{slide_path}", exc_info=e)

    return None



from __future__ import annotations

from pathlib import Path
from typing import Callable
from typing import Sequence

import h5py
import numpy as np
import numpy.typing as npt
import torch
from PIL import Image

from wsinfer.wsi import WSI


def _read_patch_coords(path: str | Path) -> npt.NDArray[np.int_]:
    """Read HDF5 file of patch coordinates are return numpy array.

    Returned array has shape (num_patches, 4). Each row has values
    [minx, miny, width, height].
    """
    with h5py.File(path, mode="r") as f:
        coords: npt.NDArray[np.int_] = f["/coords"][()]
        coords_metadata = f["/coords"].attrs
        if "patch_level" not in coords_metadata.keys():
            raise KeyError(
                "Could not find required key 'patch_level' in hdf5 of patch "
                "coordinates. Has the version of CLAM been updated?"
            )
        patch_level = coords_metadata["patch_level"]
        if patch_level != 0:
            raise NotImplementedError(
                f"This script is designed for patch_level=0 but got {patch_level}"
            )
        if coords.ndim != 2:
            raise ValueError(f"expected coords to have 2 dimensions, got {coords.ndim}")
        if coords.shape[1] != 2:
            raise ValueError(
                f"expected second dim of coords to have len 2 but got {coords.shape[1]}"
            )

        if "patch_size" not in coords_metadata.keys():
            raise KeyError("expected key 'patch_size' in attrs of coords dataset")
        # Append width and height values to the coords, so now each row is
        # [minx, miny, width, height]
        wh = np.full_like(coords, coords_metadata["patch_size"])
        coords = np.concatenate((coords, wh), axis=1)

    return coords


class WholeSlideImagePatches(torch.utils.data.Dataset):
    """Dataset of one whole slide image.

    This object retrieves patches from a whole slide image on the fly.

    Parameters
    ----------
    wsi_path : str, Path
        Path to whole slide image file.
    patch_path : str, Path
        Path to npy file with coordinates of input image.
    um_px : float
        Scale of the resulting patches. For example, 0.5 for ~20x magnification.
    patch_size : int
        The size of patches in pixels.
    transform : callable, optional
        A callable to modify a retrieved patch. The callable must accept a
        PIL.Image.Image instance and return a torch.Tensor.
    """

    def __init__(
        self,
        wsi_path: str | Path,
        patch_path: str | Path,
        um_px: float,
        patch_size: int,
        transform: Callable[[Image.Image], torch.Tensor] | None = None,
    ):
        self.wsi_path = wsi_path
        self.patch_path = patch_path
        self.um_px = float(um_px)
        self.patch_size = int(patch_size)
        self.transform = transform

        assert Path(wsi_path).exists(), "wsi path not found"
        assert Path(patch_path).exists(), "patch path not found"

        self.patches = _read_patch_coords(self.patch_path)

        assert self.patches.ndim == 2, "expected 2D array of patch coordinates"
        # x, y, width, height
        assert self.patches.shape[1] == 4, "expected second dimension to have len 4"

    def worker_init(self, worker_id: int | None = None) -> None:
        del worker_id
        self.slide = WSI(self.wsi_path)

    def __len__(self) -> int:
        return self.patches.shape[0]

    def __getitem__(self, idx: int) -> tuple[Image.Image | torch.Tensor, torch.Tensor]:
        coords: Sequence[int] = self.patches[idx]
        assert len(coords) == 4, "expected 4 coords (minx, miny, width, height)"
        minx, miny, width, height = coords

        patch_im = self.slide.read_region(
            location=(minx, miny), level=0, size=(width, height)
        )
        patch_im = patch_im.convert("RGB")

        if self.transform is not None:
            patch_im = self.transform(patch_im)

        return patch_im, torch.as_tensor([minx, miny, width, height])



from __future__ import annotations

import logging
from fractions import Fraction
from pathlib import Path
from typing import Literal
from typing import Protocol
from typing import overload

import tifffile
from PIL import Image

from .errors import CannotReadSpacing
from .errors import DuplicateFilePrefixesFound
from .errors import NoBackendException

logger = logging.getLogger(__name__)


try:
    import openslide

    HAS_OPENSLIDE = True
    logger.debug("Imported openslide")
except Exception as err:
    HAS_OPENSLIDE = False
    logger.debug(f"Unable to import openslide due to error: {err}")

try:
    import tiffslide

    HAS_TIFFSLIDE = True
    logger.debug("Imported tiffslide")
except Exception as err:
    HAS_TIFFSLIDE = False
    logger.debug(f"Unable to import tiffslide due to error: {err}")


@overload
def set_backend(name: Literal["openslide"]) -> type[openslide.OpenSlide]:
    ...


@overload
def set_backend(name: Literal["tiffslide"]) -> type[tiffslide.TiffSlide]:
    ...


def set_backend(
    name: Literal["openslide"] | Literal["tiffslide"],
) -> type[tiffslide.TiffSlide] | type[openslide.OpenSlide]:
    global WSI
    if name not in ["openslide", "tiffslide"]:
        raise ValueError(f"Unknown backend: {name}")
    logger.info(f"Setting backend to {name}")
    if name == "openslide":
        WSI = openslide.OpenSlide
    elif name == "tiffslide":
        WSI = tiffslide.TiffSlide
    else:
        raise ValueError(f"Unknown backend: {name}")
    return WSI


# Set the slide backend based on the environment.
WSI: type[openslide.OpenSlide] | type[tiffslide.TiffSlide]
if HAS_OPENSLIDE:
    WSI = set_backend("openslide")
elif HAS_TIFFSLIDE:
    WSI = set_backend("tiffslide")
else:
    raise NoBackendException("No backend found! Please install openslide or tiffslide")


# For typing an object that has a method `read_region`.


class CanReadRegion(Protocol):
    def read_region(
        self, location: tuple[int, int], level: int, size: tuple[int, int]
    ) -> Image.Image:
        pass


def _get_mpp_openslide(slide_path: str | Path) -> tuple[float, float]:
    """Read MPP using OpenSlide.

    Parameters
    ----------
    slide_path : str or Path
        The path to the whole slide image.

    Returns
    -------
    mppx, mppy
        Two floats representing the micrometers per pixel in x and y dimensions.

    Raises
    ------
    CannotReadSpacing if spacing cannot be read from the whole slide iamge.
    """
    logger.debug("Attempting to read MPP using OpenSlide")
    slide = openslide.OpenSlide(slide_path)
    mppx: float | None = None
    mppy: float | None = None

    if (
        openslide.PROPERTY_NAME_MPP_X in slide.properties
        and openslide.PROPERTY_NAME_MPP_Y in slide.properties
    ):
        logger.debug(
            "Properties of the OpenSlide object contains keys"
            f" {openslide.PROPERTY_NAME_MPP_X} and {openslide.PROPERTY_NAME_MPP_Y}"
        )
        mppx = slide.properties[openslide.PROPERTY_NAME_MPP_X]
        mppy = slide.properties[openslide.PROPERTY_NAME_MPP_Y]
        logger.debug(
            f"Value of {openslide.PROPERTY_NAME_MPP_X} is {mppx} and value"
            f" of {openslide.PROPERTY_NAME_MPP_Y} is {mppy}"
        )
        if mppx is not None and mppy is not None:
            try:
                logger.debug("Attempting to convert these MPP strings to floats")
                mppx = float(mppx)
                mppy = float(mppy)
                return mppx, mppy
            except Exception as err:
                logger.debug(f"Exception caught while converting to float: {err}")
    elif (
        "tiff.ResolutionUnit" in slide.properties
        and "tiff.XResolution" in slide.properties
        and "tiff.YResolution" in slide.properties
    ):
        logger.debug("Attempting to read spacing using openslide and tiff tags")
        resunit = slide.properties["tiff.ResolutionUnit"].lower()
        if resunit not in {"millimeter", "centimeter", "cm", "inch"}:
            raise CannotReadSpacing(f"unknown resolution unit: '{resunit}'")
        scale = {
            "inch": 25400.0,
            "centimeter": 10000.0,
            "cm": 10000.0,
            "millimeter": 1000.0,
        }.get(resunit, None)

        x_resolution = float(slide.properties["tiff.XResolution"])
        y_resolution = float(slide.properties["tiff.YResolution"])

        if scale is not None:
            try:
                mpp_x = scale / x_resolution
                mpp_y = scale / y_resolution
                return mpp_x, mpp_y
            except ArithmeticError as err:
                raise CannotReadSpacing(
                    f"error in math {scale} / {x_resolution}"
                    f" or {scale} / {y_resolution}"
                ) from err
        else:
            raise CannotReadSpacing()

    else:
        logger.debug(
            "Properties of the OpenSlide object does not contain keys"
            f" {openslide.PROPERTY_NAME_MPP_X} and {openslide.PROPERTY_NAME_MPP_Y}"
        )
    raise CannotReadSpacing()


def _get_mpp_tiffslide(
    slide_path: str | Path,
) -> tuple[float, float]:
    """Read MPP using TiffSlide."""
    slide = tiffslide.TiffSlide(slide_path)
    mppx: float | None = None
    mppy: float | None = None
    if (
        tiffslide.PROPERTY_NAME_MPP_X in slide.properties
        and tiffslide.PROPERTY_NAME_MPP_Y in slide.properties
    ):
        mppx = slide.properties[tiffslide.PROPERTY_NAME_MPP_X]
        mppy = slide.properties[tiffslide.PROPERTY_NAME_MPP_Y]
        if mppx is None or mppy is None:
            raise CannotReadSpacing()
        else:
            try:
                mppx = float(mppx)
                mppy = float(mppy)
                return mppx, mppy
            except Exception as err:
                raise CannotReadSpacing() from err
    raise CannotReadSpacing()


# Modified from
# https://github.com/bayer-science-for-a-better-life/tiffslide/blob/8bea5a4c8e1429071ade6d4c40169ce153786d19/tiffslide/tiffslide.py#L712-L745
def _get_mpp_tifffile(slide_path: str | Path) -> tuple[float, float]:
    """Read MPP using Tifffile."""
    with tifffile.TiffFile(slide_path) as tif:
        series0 = tif.series[0]
        page0 = series0[0]
        try:
            resolution_unit = page0.tags["ResolutionUnit"].value
            x_resolution = Fraction(*page0.tags["XResolution"].value)
            y_resolution = Fraction(*page0.tags["YResolution"].value)
        except KeyError as err:
            raise CannotReadSpacing() from err

        RESUNIT = tifffile.TIFF.RESUNIT
        scale = {
            RESUNIT.INCH: 25400.0,
            RESUNIT.CENTIMETER: 10000.0,
            RESUNIT.MILLIMETER: 1000.0,
            RESUNIT.MICROMETER: 1.0,
            RESUNIT.NONE: None,
        }.get(resolution_unit, None)
        if scale is not None:
            try:
                mpp_x = scale / x_resolution
                mpp_y = scale / y_resolution
                return mpp_x, mpp_y
            except ArithmeticError as err:
                raise CannotReadSpacing() from err
    raise CannotReadSpacing()


def get_avg_mpp(slide_path: Path | str) -> float:
    """Return the average MPP of a whole slide image.

    The value is in units of micrometers per pixel and is
    the average of the X and Y dimensions.

    Raises
    ------
    CannotReadSpacing if the spacing cannot be read.
    """

    mppx: float
    mppy: float

    if HAS_OPENSLIDE:
        try:
            mppx, mppy = _get_mpp_openslide(slide_path)
            return (mppx + mppy) / 2
        except CannotReadSpacing:
            # At this point, we want to continue to other implementations.
            pass
    if HAS_TIFFSLIDE:
        try:
            mppx, mppy = _get_mpp_tiffslide(slide_path)
            return (mppx + mppy) / 2
        except CannotReadSpacing:
            # Our last hope to read the mpp is tifffile.
            pass
    try:
        mppx, mppy = _get_mpp_tifffile(slide_path)
        return (mppx + mppy) / 2
    except CannotReadSpacing:
        pass

    raise CannotReadSpacing(slide_path)


def _validate_wsi_directory(wsi_dir: str | Path) -> None:
    """Validate a directory of whole slide images."""
    wsi_dir = Path(wsi_dir)
    maybe_slides = sorted(wsi_dir.glob("*"))
    uniq_stems = set(p.stem for p in maybe_slides)
    if len(uniq_stems) != len(maybe_slides):
        raise DuplicateFilePrefixesFound(
            "A slide with the same prefix but different extensions has been found"
            " (like slide.svs and slide.tif). Slides must have unique prefixes."
        )



"""Detect cancerous regions in a whole slide image."""

from __future__ import annotations

import dataclasses
import getpass
import json
import os
import platform
import shutil
import subprocess
import sys
from datetime import datetime
from pathlib import Path
from typing import Any

import click
import wsinfer_zoo
import wsinfer_zoo.client
from wsinfer_zoo.client import HFModel
from wsinfer_zoo.client import ModelConfiguration

from ..modellib import models
from ..modellib.run_inference import run_inference
from ..patchlib import segment_and_patch_directory_of_slides


def _num_cpus() -> int:
    """Get number of CPUs on the system."""
    try:
        return len(os.sched_getaffinity(0))
    # os.sched_getaffinity seems to be linux only.
    except AttributeError:
        count = os.cpu_count()  # potentially None
        return count or 0


def _inside_container() -> str:
    if Path("/.dockerenv").exists():
        return "yes, docker"
    elif (
        Path("/singularity").exists()
        or Path("/singularity.d").exists()
        or Path("/.singularity.d").exists()
    ):
        # TODO: apptainer might change the name of this directory.
        return "yes, apptainer/singularity"
    return "no"


def _get_timestamp() -> str:
    dt = datetime.now().astimezone()
    # Thu Aug 25 23:32:17 2022 EDT
    return dt.strftime("%c %Z")


def _print_system_info() -> None:
    """Print information about the system."""
    import torch
    import torchvision

    from .. import __version__

    click.secho(f"\nRunning wsinfer version {__version__}", fg="green")
    print("\nIf you run into issues, please submit a new issue at")
    print("https://github.com/kaczmarj/patch-classification-pipeline/issues/new")
    print("\nSystem information")
    print("------------------")
    print(f"Timestamp: {_get_timestamp()}")
    print(f"{platform.platform()}")
    try:
        print(f"User: {getpass.getuser()}")
    except KeyError:
        # If /etc/passwd does not include the username of the current user ID, a
        # KeyError is thrown. This could happen in a Docker image running as a different
        # user with `--user $(id -u):$(id -g)` but that does not bind mount the
        # /etc/passwd file.
        print("User: UNKNOWN")
    print(f"Hostname: {platform.node()}")
    print(f"Working directory: {os.getcwd()}")
    print(f"In container: {_inside_container()}")
    print(f"Python executable: {sys.executable}")
    print(f"Python version: {platform.python_version()}")
    print(f"  Torch version: {torch.__version__}")
    print(f"  Torchvision version: {torchvision.__version__}")
    cuda_is_available = torch.cuda.is_available()
    if cuda_is_available:
        click.secho("GPU available", fg="green")
        click.secho(f"  Using {torch.cuda.device_count()} GPU(s)", fg="green")
        cuda_ver = torch.version.cuda or "NOT FOUND"
        print(f"  CUDA version: {cuda_ver}")
    else:
        click.secho("GPU not available", bg="red", fg="black")
    cuda_visible_devices = os.environ.get("CUDA_VISIBLE_DEVICES", "NOT SET")
    print(f"CUDA_VISIBLE_DEVICES: {cuda_visible_devices}")
    if torch.version.cuda is None:
        click.secho("\n*******************************************", fg="yellow")
        click.secho("GPU WILL NOT BE USED", fg="yellow")
        if torch.version.cuda is None:
            click.secho("  CPU-only version of PyTorch is installed", fg="yellow")
        click.secho("*******************************************", fg="yellow")
    elif not cuda_is_available:
        click.secho("\n*******************************************", fg="yellow")
        click.secho("GPU WILL NOT BE USED", fg="yellow")
        if torch.version.cuda is None:
            click.secho("  CUDA DEVICES NOT AVAILABLE", fg="yellow")
        click.secho("*******************************************", fg="yellow")


def _get_info_for_save(
    model_obj: models.LocalModelTorchScript | HFModel,
) -> dict[str, Any]:
    """Get dictionary with information about the run. To save as JSON in output dir."""

    import torch

    from .. import __version__

    here = Path(__file__).parent.resolve()

    def get_git_info() -> dict[str, str | bool]:
        here = Path(__file__).parent.resolve()

        def get_stdout(args: list[str]) -> str:
            proc = subprocess.run(args, capture_output=True, cwd=here)
            return "" if proc.returncode != 0 else proc.stdout.decode().strip()

        git_remote = get_stdout("git config --get remote.origin.url".split())
        git_branch = get_stdout("git rev-parse --abbrev-ref HEAD".split())
        git_commit = get_stdout("git rev-parse HEAD".split())

        # https://stackoverflow.com/a/3879077/5666087
        cmd = subprocess.run("git diff-index --quiet HEAD --".split(), cwd=here)
        uncommitted_changes = cmd.returncode != 0
        return {
            "git_remote": git_remote,
            "git_branch": git_branch,
            "git_commit": git_commit,
            "uncommitted_changes": uncommitted_changes,
        }

    # Test if we are in a git repo. If we are, then get git info.
    git_program = shutil.which("git")
    git_installed = git_program is not None
    is_git_repo = False
    if git_installed:
        cmd = subprocess.run(
            [str(git_program), "branch"],
            cwd=here,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        is_git_repo = cmd.returncode == 0
    git_info = None
    if git_installed and is_git_repo:
        git_info = get_git_info()

    hf_info = None
    if hasattr(model_obj, "hf_info"):
        hf_info = dataclasses.asdict(model_obj.hf_info)

    return {
        "model": {
            "config": dataclasses.asdict(model_obj.config),
            "huggingface_location": hf_info,
            "path": str(model_obj.model_path),
        },
        "runtime": {
            "version": __version__,
            "working_dir": os.getcwd(),
            "args": " ".join(sys.argv),
            "python_executable": sys.executable,
            "python_version": platform.python_version(),
            "in_container": _inside_container(),
            "pytorch_version": torch.__version__,
            "cuda_version": torch.version.cuda,
            "git": git_info,
            "wsinfer_zoo_version": wsinfer_zoo.__version__,
        },
        "timestamp": _get_timestamp(),
    }


@click.command()
@click.pass_context
@click.option(
    "-i",
    "--wsi-dir",
    type=click.Path(exists=True, file_okay=False, path_type=Path, resolve_path=True),
    required=True,
    help="Directory containing whole slide images. This directory can *only* contain"
    " whole slide images.",
)
@click.option(
    "-o",
    "--results-dir",
    type=click.Path(file_okay=False, path_type=Path, resolve_path=True),
    required=True,
    help="Directory to store results. If directory exists, will skip"
    " whole slides for which outputs exist.",
)
@click.option(
    "-m",
    "--model",
    "model_name",
    type=click.Choice(sorted(wsinfer_zoo.client.load_registry().models.keys())),
    help="Name of the model to use from WSInfer Model Zoo. Mutually exclusive with"
    " --config.",
)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, dir_okay=False, path_type=Path, resolve_path=True),
    help=(
        "Path to configuration for the trained model. Use this option if the"
        " model weights are not registered in wsinfer. Mutually exclusive with"
        "--model"
    ),
)
@click.option(
    "-p",
    "--model-path",
    type=click.Path(exists=True, dir_okay=False, path_type=Path, resolve_path=True),
    help=(
        "Path to the pretrained model. Use only when --config is passed. Mutually "
        "exclusive with --model."
    ),
)
@click.option(
    "-b",
    "--batch-size",
    type=click.IntRange(min=1),
    default=32,
    show_default=True,
    help="Batch size during model inference. If using multiple GPUs, increase the"
    " batch size.",
)
@click.option(
    "-n",
    "--num-workers",
    default=min(_num_cpus(), 8),  # Use at most 8 workers by default.
    show_default=True,
    type=click.IntRange(min=0),
    help="Number of workers to use for data loading during model inference (n=0 for"
    " single thread). Set this to the number of cores on your machine or lower.",
)
@click.option(
    "--speedup/--no-speedup",
    default=False,
    show_default=True,
    help="JIT-compile the model and apply inference optimizations. This imposes a"
    " startup cost but may improve performance overall.",
)
def run(
    ctx: click.Context,
    *,
    wsi_dir: Path,
    results_dir: Path,
    model_name: str | None,
    config: Path | None,
    model_path: Path | None,
    batch_size: int,
    num_workers: int = 0,
    speedup: bool = False,
) -> None:
    """Run model inference on a directory of whole slide images.

    This command will create a tissue mask of each WSI. Then patch coordinates will be
    computed. The chosen model will be applied to each patch, and the results will be
    saved to a CSV in `RESULTS_DIR/model-output`.

    Example:

    CUDA_VISIBLE_DEVICES=0 wsinfer run --wsi-dir slides/ --results-dir results
    --model breast-tumor-resnet34.tcga-brca --batch-size 32 --num-workers 4

    To list all available models and weights, use `wsinfer-zoo ls`.
    """

    if model_name is None and config is None and model_path is None:
        raise click.UsageError(
            "one of --model or (--config and --model-path) is required."
        )
    elif (config is not None or model_path is not None) and model_name is not None:
        raise click.UsageError(
            "--config and --model-path are mutually exclusive with --model."
        )
    elif (config is not None) ^ (model_path is not None):  # XOR
        raise click.UsageError(
            "--config and --model-path must both be set if one is set."
        )

    wsi_dir = wsi_dir.resolve()
    results_dir = results_dir.resolve()

    if not wsi_dir.exists():
        raise FileNotFoundError(f"Whole slide image directory not found: {wsi_dir}")

    # Test that wsi dir actually includes files. This is here for an interesting edge
    # case. When using a Linux container and if the data directory is symlinked from a
    # different directory, both directories need to be bind mounted onto the container.
    # If only the symlinked directory is included, then the patching script will fail,
    # even though it looks like there are files in the wsi_dir directory.
    files_in_wsi_dir = [p for p in wsi_dir.glob("*") if p.exists()]
    if not files_in_wsi_dir:
        raise FileNotFoundError(f"no files exist in the slide directory: {wsi_dir}")

    _print_system_info()

    print("\nCommand line arguments")
    print("----------------------")
    for key, value in ctx.params.items():
        print(f"{key} = {value}")
    print("----------------------\n")

    # Get weights object before running the patching script because we need to get the
    # necessary spacing and patch size.
    model_obj: HFModel | models.LocalModelTorchScript
    if model_name is not None:
        model_obj = models.get_registered_model(name=model_name)
    elif config is not None:
        with open(config) as f:
            _config_dict = json.load(f)
        model_config = ModelConfiguration.from_dict(_config_dict)
        model_obj = models.LocalModelTorchScript(
            config=model_config, model_path=str(model_path)
        )
        del _config_dict, model_config
    else:
        raise click.ClickException("Neither of --config and --model was passed")

    click.secho("\nFinding patch coordinates...\n", fg="green")

    # FIXME: add presets for different tissue types?

    segment_and_patch_directory_of_slides(
        wsi_dir=wsi_dir,
        save_dir=results_dir,
        patch_size_px=model_obj.config.patch_size_pixels,
        patch_spacing_um_px=model_obj.config.spacing_um_px,
        thumbsize=(2048, 2048),
        # TODO: these can be made arguments to the CLI.
        median_filter_size=7,
        binary_threshold=7,
        closing_kernel_size=6,
        min_object_size_um2=200**2,
        min_hole_size_um2=190**2,
    )

    click.secho("\nRunning model inference.\n", fg="green")
    failed_patching, failed_inference = run_inference(
        wsi_dir=wsi_dir,
        results_dir=results_dir,
        model_info=model_obj,
        batch_size=batch_size,
        num_workers=num_workers,
        speedup=speedup,
    )

    if failed_patching:
        click.secho(f"\nPatching failed for {len(failed_patching)} slides", fg="yellow")
        click.secho("\n".join(failed_patching), fg="yellow")
    if failed_inference:
        click.secho(
            f"\nInference failed for {len(failed_inference)} slides", fg="yellow"
        )
        click.secho("\n".join(failed_inference), fg="yellow")

    timestamp = datetime.now().astimezone().strftime("%Y%m%dT%H%M%S")
    run_metadata_outpath = results_dir / f"run_metadata_{timestamp}.json"
    click.echo(f"Saving metadata about run to {run_metadata_outpath}")
    run_metadata = _get_info_for_save(model_obj)
    with open(run_metadata_outpath, "w") as f:
        json.dump(run_metadata, f, indent=2)

    click.secho("Finished.", fg="green")



"""Convert wsinfer models to torchscript."""

import hashlib
from pathlib import Path

import torch
from wsinfer import get_model_weights, list_all_models_and_weights


def sha256sum(path) -> str:
    """Calculate MD5 hash of a file."""
    sha256 = hashlib.sha256()
    with open(path, "rb") as f:
        while True:
            data = f.read(1024 * 64)  # 64 kb
            if not data:
                break
            sha256.update(data)
    return sha256.hexdigest()


def get_torchscript_frozen(arch: str, name: str):
    """Convert a wsinfer pytorch model to a frozen TorchScript model."""

    print("Loading model...")
    weights_obj = get_model_weights(architecture=arch, name=name)
    del arch, name

    width = weights_obj.patch_size_pixels
    example_input = torch.ones(1, 3, width, width)

    model = weights_obj.load_model()
    model.eval()

    print("Converting model to TorchScript...")
    try:
        model_jit = torch.jit.script(model, example_inputs=[(example_input,)])
    except RuntimeError:
        print("!!!!!!")
        print("torch.jit.script failed!")
    try:
        model_jit = torch.jit.trace(model, example_input)
    except Exception:
        return None, None
    model_jit_frozen = torch.jit.freeze(model_jit)

    print("Validating that jit model gives same output as original...")
    with torch.no_grad():
        output: torch.Tensor = model(example_input)
        output_jit: torch.Tensor = model_jit(example_input)
        assert torch.allclose(output, output_jit)
        output_jit_frozen: torch.Tensor = model_jit_frozen(example_input)
        assert torch.allclose(output, output_jit_frozen)

    return model_jit_frozen, weights_obj


if __name__ == "__main__":
    outdir = Path("jit-frozen-models")
    outdir.mkdir(exist_ok=True)

    for model, weights in list_all_models_and_weights():
        print(f"{model} with {weights} weights")
        frozen_model, w = get_torchscript_frozen(model, weights)
        if frozen_model is None:
            print("Failed to convert.")
            print("-" * 40)
            continue
        p = Path(w.url_file_name)
        p = outdir / f"{p.stem}-jit-frozen.pt"
        print(f"Saving to {p}")
        torch.jit.save(frozen_model, p)
        print("Calculating sha256 hash")
        sha = sha256sum(p)
        newname = p.parent / f"{p.stem}-{sha[:8]}.pt"
        print(f"Renaming to {newname}")
        p.rename(newname)
        print("-" * 40)



"""Convert VGG16 weights from TensorFlow (1.x) format to PyTorch.

The input path is _not_ the actual path to the checkpoint files. It is the stem
of these files. For example, if the checkpoint files are "foo.ckpt.index", pass the
name "foo.ckpt" to this script.

This script requires TensorFlow 2.x. Install it with 'pip install tensorflow-cpu'.

The original TIL model is at
https://stonybrookmedicine.app.box.com/v/til-results-new-model/folder/128593971923.

It was implemented in TF Slim. This script converts those TF Slim weights to PyTorch.
"""

import argparse
import os
from pathlib import Path

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"  # Do not print INFO messages

import tensorflow as tf
import torch
import torchvision

if not tf.__version__.startswith("2."):
    raise EnvironmentError("TensorFlow 2.x must be installed.")

tf_to_pytorch_layers = [
    ("vgg_16/conv1/conv1_1", "features.0"),
    ("vgg_16/conv1/conv1_2", "features.2"),
    ("vgg_16/conv2/conv2_1", "features.5"),
    ("vgg_16/conv2/conv2_2", "features.7"),
    ("vgg_16/conv3/conv3_1", "features.10"),
    ("vgg_16/conv3/conv3_2", "features.12"),
    ("vgg_16/conv3/conv3_3", "features.14"),
    ("vgg_16/conv4/conv4_1", "features.17"),
    ("vgg_16/conv4/conv4_2", "features.19"),
    ("vgg_16/conv4/conv4_3", "features.21"),
    ("vgg_16/conv5/conv5_1", "features.24"),
    ("vgg_16/conv5/conv5_2", "features.26"),
    ("vgg_16/conv5/conv5_3", "features.28"),
    ("vgg_16/fc6", "classifier.0"),
    ("vgg_16/fc7", "classifier.3"),
    ("vgg_16/fc8", "classifier.6"),
]


def convert_tf_to_pytorch(input_path, output_path, num_classes: int):
    try:
        ckpt = tf.train.load_checkpoint(input_path)
    except tf.errors.DataLossError:
        raise RuntimeError(
            "Error: could not load checkpoint. Did you pass in the stem of the path?"
            "Pass in the path without '.index' or '.meta' or '.data-00000-of-00001'."
        )
    new_state_dict = {}
    for tf_prefix, torch_prefix in tf_to_pytorch_layers:
        tf_weights = f"{tf_prefix}/weights"
        tf_biases = f"{tf_prefix}/biases"
        torch_weights = f"{torch_prefix}.weight"
        torch_biases = f"{torch_prefix}.bias"
        tf_weight_array = ckpt.get_tensor(tf_weights)
        tf_bias_array = ckpt.get_tensor(tf_biases)
        if "conv" in tf_weights:
            tf_weight_array = tf_weight_array.transpose([3, 2, 0, 1])
        elif "fc" in tf_weights:
            if tf_weights == "vgg_16/fc6/weights":
                # [7, 7, 512, 4096] -> [25088, 4096]
                tf_weight_array = tf_weight_array.transpose([2, 0, 1, 3])
                tf_weight_array = tf_weight_array.reshape((25088, 4096))
            # E.g., go from shape [1, 1, 4096, 1000] to [1000, 4096]
            tf_weight_array = tf_weight_array.squeeze().T
        new_state_dict[torch_weights] = torch.from_numpy(tf_weight_array)
        new_state_dict[torch_biases] = torch.from_numpy(tf_bias_array)

    # Test that conversion was (probably) done correctly.
    true_model = torchvision.models.vgg16()
    true_model.classifier[6] = torch.nn.Linear(4096, num_classes)
    if true_model.state_dict().keys() != new_state_dict.keys():
        raise RuntimeError(
            "Something went wrong... converted model keys do not match TorchVision"
            " VGG16 keys."
        )
    true_state_dict = true_model.state_dict()
    for true_k, new_k in zip(true_state_dict, new_state_dict):
        true_shape = true_state_dict[true_k].shape
        new_shape = new_state_dict[new_k].shape
        if true_shape != new_shape:
            raise ValueError(
                "Shape mismatch between converted parameters and reference parameters."
            )

    torch.save(new_state_dict, output_path)


if __name__ == "__main__":
    p = argparse.ArgumentParser(description=__doc__)
    p.add_argument("input", help="Path to the VGG16 tensorflow checkpoint.")
    p.add_argument("output")
    p.add_argument(
        "--num-classes", type=int, default=2, help="Number of output classes."
    )
    args = p.parse_args()
    # We do not expect the input "path" to exist, because it should be the stem.
    if Path(args.input).exists():
        raise ValueError(
            "Input path exists. Instead of passing the full path, pass in the stem"
            " (i.e., without '.index' or '.meta' or '.data-00000-of-00001'."
        )
    convert_tf_to_pytorch(args.input, args.output, args.num_classes)



"""Run inference.

From the original paper (https://www.ncbi.nlm.nih.gov/pmc/articles/PMC7369575/):
> In the prediction (test) phase, no data augmentation was applied except for the
> normalization of the color channels.
"""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from typing import cast as type_cast

import numpy as np
import numpy.typing as npt
import pandas as pd
import torch
import tqdm
import wsinfer_zoo.client

from .. import errors
from ..wsi import _validate_wsi_directory
from .data import WholeSlideImagePatches
from .models import LocalModelTorchScript
from .models import get_pretrained_torch_module
from .models import jit_compile
from .transforms import make_compose_from_transform_config


def run_inference(
    wsi_dir: str | Path,
    results_dir: str | Path,
    model_info: wsinfer_zoo.client.HFModelTorchScript | LocalModelTorchScript,
    batch_size: int = 32,
    num_workers: int = 0,
    speedup: bool = False,
) -> tuple[list[str], list[str]]:
    """Run model inference on a directory of whole slide images and save results to CSV.

    This assumes the patching has already been done and the results are stored in
    `results_dir`. An error will be raised otherwise.

    Output CSV files are written to `{results_dir}/model-outputs/`.

    Parameters
    ----------
    wsi_dir : str or Path
        Directory containing whole slide images. This directory can *only* contain
        whole slide images. Otherwise, an error will be raised during model inference.
    results_dir : str or Path
        Directory containing results of patching.
    model_info :
        Instance of Weights including the model object and information about how to
        apply the model to new data.
    batch_size : int
        The batch size during the forward pass (default is 32).
    num_workers : int
        Number of workers for data loading (default is 0, meaning use a single thread).
    speedup : bool
        If True, JIT-compile the model. This has a startup cost but model inference
        should be faster (default False).

    Returns
    -------
    A tuple of two lists of strings. The first list contains the slide IDs for which
    patching failed, and the second list contains the slide IDs for which model
    inference failed.
    """
    # Make sure required directories exist.
    wsi_dir = Path(wsi_dir)
    if not wsi_dir.exists():
        raise errors.WholeSlideImageDirectoryNotFound(f"directory not found: {wsi_dir}")
    wsi_paths = list(wsi_dir.glob("*"))
    if not wsi_paths:
        raise errors.WholeSlideImagesNotFound(wsi_dir)
    results_dir = Path(results_dir)
    if not results_dir.exists():
        raise errors.ResultsDirectoryNotFound(results_dir)

    _validate_wsi_directory(wsi_dir)

    # Check patches directory.
    patch_dir = results_dir / "patches"
    if not patch_dir.exists():
        raise errors.PatchDirectoryNotFound("Results dir must include 'patches' dir")
    # Create the patch paths based on the whole slide image paths. In effect, only
    # create patch paths if the whole slide image patch exists.
    patch_paths = [patch_dir / p.with_suffix(".h5").name for p in wsi_paths]

    model_output_dir = results_dir / "model-outputs"
    model_output_dir.mkdir(exist_ok=True)

    model = get_pretrained_torch_module(model=model_info)
    model.eval()

    # Set the device.
    if torch.cuda.is_available():
        device = torch.device("cuda")
        if torch.cuda.device_count() > 1:
            model = torch.nn.DataParallel(model)
    elif torch.backends.mps.is_available() and torch.backends.mps.is_built():
        device = torch.device("mps")
    else:
        device = torch.device("cpu")
    print(f'Using device "{device}"')

    model.to(device)

    if speedup:
        if TYPE_CHECKING:
            model = type_cast(torch.nn.Module, jit_compile(model))
        else:
            model = jit_compile(model)

    transform = make_compose_from_transform_config(model_info.config.transform)

    failed_patching = [p.stem for p in patch_paths if not p.exists()]
    failed_inference: list[str] = []

    for i, (wsi_path, patch_path) in enumerate(zip(wsi_paths, patch_paths)):
        print(f"Slide {i+1} of {len(wsi_paths)}")
        print(f" Slide path: {wsi_path}")
        print(f" Patch path: {patch_path}")

        slide_csv_name = Path(wsi_path).with_suffix(".csv").name
        slide_csv = model_output_dir / slide_csv_name
        if slide_csv.exists():
            print("Output CSV exists... skipping.")
            print(slide_csv)
            continue

        if not patch_path.exists():
            print(f"Skipping because patch file not found: {patch_path}")
            continue

        try:
            dset = WholeSlideImagePatches(
                wsi_path=wsi_path,
                patch_path=patch_path,
                um_px=model_info.config.spacing_um_px,
                patch_size=model_info.config.patch_size_pixels,
                transform=transform,
            )
        except Exception:
            failed_inference.append(wsi_dir.stem)
            continue

        # The worker_init_fn does not seem to be used when num_workers=0
        # so we call it manually to finish setting up the dataset.
        if num_workers == 0:
            dset.worker_init()

        loader = torch.utils.data.DataLoader(
            dset,
            batch_size=batch_size,
            shuffle=False,
            num_workers=num_workers,
            worker_init_fn=dset.worker_init,
        )

        # Store the coordinates and model probabiltiies of each patch in this slide.
        # This lets us know where the probabiltiies map to in the slide.
        slide_coords: list[npt.NDArray[np.int_]] = []
        slide_probs: list[npt.NDArray[np.float_]] = []
        for batch_imgs, batch_coords in tqdm.tqdm(loader):
            assert batch_imgs.shape[0] == batch_coords.shape[0], "length mismatch"
            with torch.no_grad():
                logits: torch.Tensor = model(batch_imgs.to(device)).detach().cpu()
            # probs has shape (batch_size, num_classes) or (batch_size,)
            if len(logits.shape) > 1 and logits.shape[1] > 1:
                probs = torch.nn.functional.softmax(logits, dim=1)
            else:
                probs = torch.sigmoid(logits.squeeze(1))
            slide_coords.append(batch_coords.numpy())
            slide_probs.append(probs.numpy())

        slide_coords_arr = np.concatenate(slide_coords, axis=0)
        slide_df = pd.DataFrame(
            dict(
                minx=slide_coords_arr[:, 0],
                miny=slide_coords_arr[:, 1],
                width=slide_coords_arr[:, 2],
                height=slide_coords_arr[:, 3],
            )
        )
        slide_probs_arr = np.concatenate(slide_probs, axis=0)
        # Use 'prob-' prefix for all classes. This should make it clearer that the
        # column has probabilities for the class. It also makes it easier for us to
        # identify columns associated with probabilities.
        prob_colnames = [f"prob_{c}" for c in model_info.config.class_names]
        slide_df.loc[:, prob_colnames] = slide_probs_arr
        slide_df.to_csv(slide_csv, index=False)
        print("-" * 40)

    return failed_patching, failed_inference



"""WSInfer is a toolkit for fast patch-based inference on whole slide images."""

from __future__ import annotations

try:
    from ._version import __version__
except ImportError:
    __version__ = "0.0.unknown"


# Patch Zarr. See:
# https://github.com/bayer-science-for-a-better-life/tiffslide/issues/72#issuecomment-1627918238
# https://github.com/zarr-developers/zarr-python/pull/1454
def _patch_zarr_kvstore() -> None:
    from zarr.storage import KVStore

    def _zarr_KVStore___contains__(self, key):  # type: ignore
        return key in self._mutable_mapping

    if "__contains__" not in KVStore.__dict__:
        KVStore.__contains__ = _zarr_KVStore___contains__


_patch_zarr_kvstore()



"""Convert CSVs of model outputs to GeoJSON files.

GeoJSON files can be loaded into whole slide image viewers like QuPath.
"""

from __future__ import annotations

import json
from pathlib import Path

import click
import pandas as pd
import tqdm


def _box_to_polygon(
    *, minx: int, miny: int, width: int, height: int
) -> list[tuple[int, int]]:
    """Get coordinates of a box polygon."""
    maxx = minx + width
    maxy = miny + height
    return [(maxx, miny), (maxx, maxy), (minx, maxy), (minx, miny), (maxx, miny)]


def _row_to_geojson(row: pd.Series, prob_cols: list[str]) -> dict:
    """Convert information about one tile to a single GeoJSON feature."""
    minx, miny, width, height = row["minx"], row["miny"], row["width"], row["height"]
    coords = _box_to_polygon(minx=minx, miny=miny, width=width, height=height)
    prob_dict = row[prob_cols].to_dict()
    measurements = [{"name": k, "value": v} for k, v in prob_dict.items()]
    return {
        "type": "Feature",
        "id": "PathTileObject",
        "geometry": {
            "type": "Polygon",
            "coordinates": [coords],
        },
        "properties": {
            "isLocked": True,
            # measurements is a list of {"name": str, "value": float} dicts.
            # https://qupath.github.io/javadoc/docs/qupath/lib/measurements/MeasurementList.html
            "measurements": measurements,
            # classification is a dict of "name": str and optionally "color": int.
            # https://qupath.github.io/javadoc/docs/qupath/lib/objects/classes/PathClass.html
            # We do not include classification because we do not enforce a single class
            # per tile.
            # "classification": {"name": class_name},
        },
    }


def _dataframe_to_geojson(df: pd.DataFrame, prob_cols: list[str]) -> dict:
    """Convert a dataframe of tiles to GeoJSON format."""
    features = df.apply(_row_to_geojson, axis=1, prob_cols=prob_cols)
    return {
        "type": "FeatureCollection",
        "features": features.tolist(),
    }


def convert(input: str | Path, output: str | Path) -> None:
    df = pd.read_csv(input)
    prob_cols = [col for col in df.columns.tolist() if col.startswith("prob_")]
    if not prob_cols:
        raise click.ClickException("Did not find any columns with prob_ prefix.")
    geojson = _dataframe_to_geojson(df, prob_cols)
    with open(output, "w") as f:
        json.dump(geojson, f)


@click.command()
@click.argument(
    "results_dir",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, path_type=Path, resolve_path=True
    ),
)
@click.argument(
    "output",
    type=click.Path(exists=False, path_type=Path, resolve_path=True),
)
def togeojson(*, results_dir: Path, output: Path) -> None:
    """Convert model outputs to GeoJSON format.

    GeoJSON files can be used with pathology viewers like QuPath.

    RESULTS_DIR     Path to results directory (containing model-outputs dir).

    OUTPUT          Path to output directory in which to save GeoJSON files.
    """
    if not results_dir.exists():
        raise click.ClickException(f"results_dir does not exist: {results_dir}")
    if output.exists():
        raise click.ClickException("Output directory already exists.")
    if (
        not (results_dir / "model-outputs").exists()
        and (results_dir / "patches").exists()
    ):
        raise click.ClickException(
            "Model outputs have not been generated yet. Please run model inference."
        )
    if not (results_dir / "model-outputs").exists():
        raise click.ClickException(
            "Expected results_dir to contain a 'model-outputs' directory but it does"
            " not. Please provide the path to the directory that contains"
            " model-outputs, masks, and patches."
        )

    csvs = list((results_dir / "model-outputs").glob("*.csv"))
    if not csvs:
        raise click.ClickException("No CSVs found. Did you generate model outputs?")

    output.mkdir(exist_ok=False)

    for input_csv in tqdm.tqdm(csvs):
        output_path = output / input_csv.with_suffix(".json").name
        convert(input=input_csv, output=output_path)

    click.secho(f"Saved outputs to {output}", fg="green")



"""Segment thumbnail of a whole slide image."""

from __future__ import annotations

import cv2 as cv
import numpy as np
import numpy.typing as npt
from skimage.morphology import binary_closing
from skimage.morphology import remove_small_holes
from skimage.morphology import remove_small_objects


def segment_tissue(
    im_arr: npt.NDArray[np.uint8],
    median_filter_size: int = 7,
    binary_threshold: int = 7,
    closing_kernel_size: int = 6,
    min_object_size_px: int = 512,
    min_hole_size_px: int = 1024,
) -> npt.NDArray[np.bool_]:
    """Create a binary tissue mask from an image.

    Parameters
    ----------
    im_arr : array-like
        RGB image array (uint8) with shape (rows, cols, 3).
    median_filter_size : int
        The kernel size for median filtering. Must be odd and greater than one.
    binary_threshold : int
        The pixel threshold for image binarization.
    closing_kernel_size : int
        The kernel size for morphological closing (in pixel units).
    min_object_size_px : int
        The minimum area of an object in pixels. If an object is smaller than this area,
        it is removed and is made into background.
    min_hole_size_px : int
        The minimum area of a hole in pixels. If a hole is smaller than this area, it is
        filled and is made into foreground.

    Returns
    -------
    mask
        Boolean array, where True values indicate presence of tissue.
    """
    im_arr = np.asarray(im_arr)
    assert im_arr.ndim == 3
    assert im_arr.shape[2] == 3

    # Convert to HSV color space.
    im_arr = cv.cvtColor(im_arr, cv.COLOR_RGB2HSV)
    im_arr = im_arr[:, :, 1]  # Keep saturation channel only.

    # Use median blurring to smooth the image.
    if median_filter_size <= 1 or median_filter_size % 2 == 0:
        raise ValueError(
            "median_filter_size must be greater than 1 and odd, but got"
            f" {median_filter_size}"
        )

    # We use opencv here instead of PIL because opencv is _much_ faster. We use skimage
    # further down for artifact removal (hole filling, object removal) because skimage
    # provides easy to use methods for those.
    im_arr = cv.medianBlur(im_arr, median_filter_size)

    # Binarize image.
    _, im_arr = cv.threshold(
        im_arr, thresh=binary_threshold, maxval=255, type=cv.THRESH_BINARY
    )

    # Convert to boolean dtype. This helps with static type analysis because at this
    # point, im_arr is a uint8 array.
    im_arr_binary = im_arr > 0

    # Closing. This removes small holes. It might not be entirely necessary because
    # we have hole removal below.
    im_arr_binary = binary_closing(
        im_arr_binary, footprint=np.ones((closing_kernel_size, closing_kernel_size))
    )

    # Remove small objects.
    im_arr_binary = remove_small_objects(im_arr_binary, min_size=min_object_size_px)

    # Remove small holes.
    im_arr_binary = remove_small_holes(im_arr_binary, area_threshold=min_hole_size_px)

    return im_arr_binary



"""Exceptions used in WSInfer."""

from __future__ import annotations


class WsinferException(Exception):
    """Base class for wsinfer exceptions."""


class UnknownArchitectureError(WsinferException):
    """Architecture is unknown and cannot be found."""


class WholeSlideImageDirectoryNotFound(WsinferException, FileNotFoundError):
    ...


class DuplicateFilePrefixesFound(WsinferException):
    """A duplicate file prefix has been found.

    An example of duplicate file prefixes is files a.svs and a.tif. WSInfer relies on
    the stems as a unique ID, so we cannot allow duplicate stems.
    """


class WholeSlideImagesNotFound(WsinferException, FileNotFoundError):
    ...


class ResultsDirectoryNotFound(WsinferException, FileNotFoundError):
    ...


class PatchDirectoryNotFound(WsinferException, FileNotFoundError):
    ...


class CannotReadSpacing(WsinferException):
    ...


class NoBackendException(WsinferException):
    ...



"""Convert InceptionV4 weights from TensorFlow (1.x) format to PyTorch.

The input path is _not_ the actual path to the checkpoint files. It is the stem
of these files. For example, if the checkpoint files are "foo.ckpt.index", pass the
name "foo.ckpt" to this script.

This script requires TensorFlow 2.x. Install it with 'pip install tensorflow-cpu'.

The original TIL model is at
https://stonybrookmedicine.app.box.com/v/til-results-new-model/folder/128593971923.

It was implemented in TF Slim. This script converts those TF Slim weights to PyTorch.
"""

import argparse
import os
from pathlib import Path

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"  # Do not print INFO messages

import tensorflow as tf
import torch

inceptionv4 = None
try:
    from wsinfer.modellib.inceptionv4_no_batchnorm import inceptionv4
except ImportError:
    pass

if not tf.__version__.startswith("2."):
    raise EnvironmentError("TensorFlow 2.x must be installed.")

tf_to_pytorch_layers = [
    # Basic convs
    ("InceptionV4/Conv2d_1a_3x3", "features.0.conv"),
    ("InceptionV4/Conv2d_2a_3x3", "features.1.conv"),
    ("InceptionV4/Conv2d_2b_3x3", "features.2.conv"),
    # Mixed_3a
    ("InceptionV4/Mixed_3a/Branch_1/Conv2d_0a_3x3", "features.3.conv.conv"),
    # Mixed_4a
    ("InceptionV4/Mixed_4a/Branch_0/Conv2d_0a_1x1", "features.4.branch0.0.conv"),
    ("InceptionV4/Mixed_4a/Branch_0/Conv2d_1a_3x3", "features.4.branch0.1.conv"),
    ("InceptionV4/Mixed_4a/Branch_1/Conv2d_0a_1x1", "features.4.branch1.0.conv"),
    ("InceptionV4/Mixed_4a/Branch_1/Conv2d_0b_1x7", "features.4.branch1.1.conv"),
    ("InceptionV4/Mixed_4a/Branch_1/Conv2d_0c_7x1", "features.4.branch1.2.conv"),
    ("InceptionV4/Mixed_4a/Branch_1/Conv2d_1a_3x3", "features.4.branch1.3.conv"),
    # Mixed_5a
    ("InceptionV4/Mixed_5a/Branch_0/Conv2d_1a_3x3", "features.5.conv.conv"),
    # InceptionA (1)
    ("InceptionV4/Mixed_5b/Branch_0/Conv2d_0a_1x1", "features.6.branch0.conv"),
    ("InceptionV4/Mixed_5b/Branch_1/Conv2d_0a_1x1", "features.6.branch1.0.conv"),
    ("InceptionV4/Mixed_5b/Branch_1/Conv2d_0b_3x3", "features.6.branch1.1.conv"),
    ("InceptionV4/Mixed_5b/Branch_2/Conv2d_0a_1x1", "features.6.branch2.0.conv"),
    ("InceptionV4/Mixed_5b/Branch_2/Conv2d_0b_3x3", "features.6.branch2.1.conv"),
    ("InceptionV4/Mixed_5b/Branch_2/Conv2d_0c_3x3", "features.6.branch2.2.conv"),
    ("InceptionV4/Mixed_5b/Branch_3/Conv2d_0b_1x1", "features.6.branch3.1.conv"),
    # InceptionA (2)
    ("InceptionV4/Mixed_5c/Branch_0/Conv2d_0a_1x1", "features.7.branch0.conv"),
    ("InceptionV4/Mixed_5c/Branch_1/Conv2d_0a_1x1", "features.7.branch1.0.conv"),
    ("InceptionV4/Mixed_5c/Branch_1/Conv2d_0b_3x3", "features.7.branch1.1.conv"),
    ("InceptionV4/Mixed_5c/Branch_2/Conv2d_0a_1x1", "features.7.branch2.0.conv"),
    ("InceptionV4/Mixed_5c/Branch_2/Conv2d_0b_3x3", "features.7.branch2.1.conv"),
    ("InceptionV4/Mixed_5c/Branch_2/Conv2d_0c_3x3", "features.7.branch2.2.conv"),
    ("InceptionV4/Mixed_5c/Branch_3/Conv2d_0b_1x1", "features.7.branch3.1.conv"),
    # InceptionA (3)
    ("InceptionV4/Mixed_5d/Branch_0/Conv2d_0a_1x1", "features.8.branch0.conv"),
    ("InceptionV4/Mixed_5d/Branch_1/Conv2d_0a_1x1", "features.8.branch1.0.conv"),
    ("InceptionV4/Mixed_5d/Branch_1/Conv2d_0b_3x3", "features.8.branch1.1.conv"),
    ("InceptionV4/Mixed_5d/Branch_2/Conv2d_0a_1x1", "features.8.branch2.0.conv"),
    ("InceptionV4/Mixed_5d/Branch_2/Conv2d_0b_3x3", "features.8.branch2.1.conv"),
    ("InceptionV4/Mixed_5d/Branch_2/Conv2d_0c_3x3", "features.8.branch2.2.conv"),
    ("InceptionV4/Mixed_5d/Branch_3/Conv2d_0b_1x1", "features.8.branch3.1.conv"),
    # InceptionA (4)
    ("InceptionV4/Mixed_5e/Branch_0/Conv2d_0a_1x1", "features.9.branch0.conv"),
    ("InceptionV4/Mixed_5e/Branch_1/Conv2d_0a_1x1", "features.9.branch1.0.conv"),
    ("InceptionV4/Mixed_5e/Branch_1/Conv2d_0b_3x3", "features.9.branch1.1.conv"),
    ("InceptionV4/Mixed_5e/Branch_2/Conv2d_0a_1x1", "features.9.branch2.0.conv"),
    ("InceptionV4/Mixed_5e/Branch_2/Conv2d_0b_3x3", "features.9.branch2.1.conv"),
    ("InceptionV4/Mixed_5e/Branch_2/Conv2d_0c_3x3", "features.9.branch2.2.conv"),
    ("InceptionV4/Mixed_5e/Branch_3/Conv2d_0b_1x1", "features.9.branch3.1.conv"),
    # ReductionA (Mixed_6a)
    ("InceptionV4/Mixed_6a/Branch_0/Conv2d_1a_3x3", "features.10.branch0.conv"),
    ("InceptionV4/Mixed_6a/Branch_1/Conv2d_0a_1x1", "features.10.branch1.0.conv"),
    ("InceptionV4/Mixed_6a/Branch_1/Conv2d_0b_3x3", "features.10.branch1.1.conv"),
    ("InceptionV4/Mixed_6a/Branch_1/Conv2d_1a_3x3", "features.10.branch1.2.conv"),
    # InceptionB (1)
    ("InceptionV4/Mixed_6b/Branch_0/Conv2d_0a_1x1", "features.11.branch0.conv"),
    ("InceptionV4/Mixed_6b/Branch_1/Conv2d_0a_1x1", "features.11.branch1.0.conv"),
    ("InceptionV4/Mixed_6b/Branch_1/Conv2d_0b_1x7", "features.11.branch1.1.conv"),
    ("InceptionV4/Mixed_6b/Branch_1/Conv2d_0c_7x1", "features.11.branch1.2.conv"),
    ("InceptionV4/Mixed_6b/Branch_2/Conv2d_0a_1x1", "features.11.branch2.0.conv"),
    ("InceptionV4/Mixed_6b/Branch_2/Conv2d_0b_7x1", "features.11.branch2.1.conv"),
    ("InceptionV4/Mixed_6b/Branch_2/Conv2d_0c_1x7", "features.11.branch2.2.conv"),
    ("InceptionV4/Mixed_6b/Branch_2/Conv2d_0d_7x1", "features.11.branch2.3.conv"),
    ("InceptionV4/Mixed_6b/Branch_2/Conv2d_0e_1x7", "features.11.branch2.4.conv"),
    ("InceptionV4/Mixed_6b/Branch_3/Conv2d_0b_1x1", "features.11.branch3.1.conv"),
    # InceptionB (2)
    ("InceptionV4/Mixed_6c/Branch_0/Conv2d_0a_1x1", "features.12.branch0.conv"),
    ("InceptionV4/Mixed_6c/Branch_1/Conv2d_0a_1x1", "features.12.branch1.0.conv"),
    ("InceptionV4/Mixed_6c/Branch_1/Conv2d_0b_1x7", "features.12.branch1.1.conv"),
    ("InceptionV4/Mixed_6c/Branch_1/Conv2d_0c_7x1", "features.12.branch1.2.conv"),
    ("InceptionV4/Mixed_6c/Branch_2/Conv2d_0a_1x1", "features.12.branch2.0.conv"),
    ("InceptionV4/Mixed_6c/Branch_2/Conv2d_0b_7x1", "features.12.branch2.1.conv"),
    ("InceptionV4/Mixed_6c/Branch_2/Conv2d_0c_1x7", "features.12.branch2.2.conv"),
    ("InceptionV4/Mixed_6c/Branch_2/Conv2d_0d_7x1", "features.12.branch2.3.conv"),
    ("InceptionV4/Mixed_6c/Branch_2/Conv2d_0e_1x7", "features.12.branch2.4.conv"),
    ("InceptionV4/Mixed_6c/Branch_3/Conv2d_0b_1x1", "features.12.branch3.1.conv"),
    # InceptionB (3)
    ("InceptionV4/Mixed_6d/Branch_0/Conv2d_0a_1x1", "features.13.branch0.conv"),
    ("InceptionV4/Mixed_6d/Branch_1/Conv2d_0a_1x1", "features.13.branch1.0.conv"),
    ("InceptionV4/Mixed_6d/Branch_1/Conv2d_0b_1x7", "features.13.branch1.1.conv"),
    ("InceptionV4/Mixed_6d/Branch_1/Conv2d_0c_7x1", "features.13.branch1.2.conv"),
    ("InceptionV4/Mixed_6d/Branch_2/Conv2d_0a_1x1", "features.13.branch2.0.conv"),
    ("InceptionV4/Mixed_6d/Branch_2/Conv2d_0b_7x1", "features.13.branch2.1.conv"),
    ("InceptionV4/Mixed_6d/Branch_2/Conv2d_0c_1x7", "features.13.branch2.2.conv"),
    ("InceptionV4/Mixed_6d/Branch_2/Conv2d_0d_7x1", "features.13.branch2.3.conv"),
    ("InceptionV4/Mixed_6d/Branch_2/Conv2d_0e_1x7", "features.13.branch2.4.conv"),
    ("InceptionV4/Mixed_6d/Branch_3/Conv2d_0b_1x1", "features.13.branch3.1.conv"),
    # InceptionB (4)
    ("InceptionV4/Mixed_6e/Branch_0/Conv2d_0a_1x1", "features.14.branch0.conv"),
    ("InceptionV4/Mixed_6e/Branch_1/Conv2d_0a_1x1", "features.14.branch1.0.conv"),
    ("InceptionV4/Mixed_6e/Branch_1/Conv2d_0b_1x7", "features.14.branch1.1.conv"),
    ("InceptionV4/Mixed_6e/Branch_1/Conv2d_0c_7x1", "features.14.branch1.2.conv"),
    ("InceptionV4/Mixed_6e/Branch_2/Conv2d_0a_1x1", "features.14.branch2.0.conv"),
    ("InceptionV4/Mixed_6e/Branch_2/Conv2d_0b_7x1", "features.14.branch2.1.conv"),
    ("InceptionV4/Mixed_6e/Branch_2/Conv2d_0c_1x7", "features.14.branch2.2.conv"),
    ("InceptionV4/Mixed_6e/Branch_2/Conv2d_0d_7x1", "features.14.branch2.3.conv"),
    ("InceptionV4/Mixed_6e/Branch_2/Conv2d_0e_1x7", "features.14.branch2.4.conv"),
    ("InceptionV4/Mixed_6e/Branch_3/Conv2d_0b_1x1", "features.14.branch3.1.conv"),
    # InceptionB (5)
    ("InceptionV4/Mixed_6f/Branch_0/Conv2d_0a_1x1", "features.15.branch0.conv"),
    ("InceptionV4/Mixed_6f/Branch_1/Conv2d_0a_1x1", "features.15.branch1.0.conv"),
    ("InceptionV4/Mixed_6f/Branch_1/Conv2d_0b_1x7", "features.15.branch1.1.conv"),
    ("InceptionV4/Mixed_6f/Branch_1/Conv2d_0c_7x1", "features.15.branch1.2.conv"),
    ("InceptionV4/Mixed_6f/Branch_2/Conv2d_0a_1x1", "features.15.branch2.0.conv"),
    ("InceptionV4/Mixed_6f/Branch_2/Conv2d_0b_7x1", "features.15.branch2.1.conv"),
    ("InceptionV4/Mixed_6f/Branch_2/Conv2d_0c_1x7", "features.15.branch2.2.conv"),
    ("InceptionV4/Mixed_6f/Branch_2/Conv2d_0d_7x1", "features.15.branch2.3.conv"),
    ("InceptionV4/Mixed_6f/Branch_2/Conv2d_0e_1x7", "features.15.branch2.4.conv"),
    ("InceptionV4/Mixed_6f/Branch_3/Conv2d_0b_1x1", "features.15.branch3.1.conv"),
    # InceptionB (6)
    ("InceptionV4/Mixed_6g/Branch_0/Conv2d_0a_1x1", "features.16.branch0.conv"),
    ("InceptionV4/Mixed_6g/Branch_1/Conv2d_0a_1x1", "features.16.branch1.0.conv"),
    ("InceptionV4/Mixed_6g/Branch_1/Conv2d_0b_1x7", "features.16.branch1.1.conv"),
    ("InceptionV4/Mixed_6g/Branch_1/Conv2d_0c_7x1", "features.16.branch1.2.conv"),
    ("InceptionV4/Mixed_6g/Branch_2/Conv2d_0a_1x1", "features.16.branch2.0.conv"),
    ("InceptionV4/Mixed_6g/Branch_2/Conv2d_0b_7x1", "features.16.branch2.1.conv"),
    ("InceptionV4/Mixed_6g/Branch_2/Conv2d_0c_1x7", "features.16.branch2.2.conv"),
    ("InceptionV4/Mixed_6g/Branch_2/Conv2d_0d_7x1", "features.16.branch2.3.conv"),
    ("InceptionV4/Mixed_6g/Branch_2/Conv2d_0e_1x7", "features.16.branch2.4.conv"),
    ("InceptionV4/Mixed_6g/Branch_3/Conv2d_0b_1x1", "features.16.branch3.1.conv"),
    # InceptionB (7)
    ("InceptionV4/Mixed_6h/Branch_0/Conv2d_0a_1x1", "features.17.branch0.conv"),
    ("InceptionV4/Mixed_6h/Branch_1/Conv2d_0a_1x1", "features.17.branch1.0.conv"),
    ("InceptionV4/Mixed_6h/Branch_1/Conv2d_0b_1x7", "features.17.branch1.1.conv"),
    ("InceptionV4/Mixed_6h/Branch_1/Conv2d_0c_7x1", "features.17.branch1.2.conv"),
    ("InceptionV4/Mixed_6h/Branch_2/Conv2d_0a_1x1", "features.17.branch2.0.conv"),
    ("InceptionV4/Mixed_6h/Branch_2/Conv2d_0b_7x1", "features.17.branch2.1.conv"),
    ("InceptionV4/Mixed_6h/Branch_2/Conv2d_0c_1x7", "features.17.branch2.2.conv"),
    ("InceptionV4/Mixed_6h/Branch_2/Conv2d_0d_7x1", "features.17.branch2.3.conv"),
    ("InceptionV4/Mixed_6h/Branch_2/Conv2d_0e_1x7", "features.17.branch2.4.conv"),
    ("InceptionV4/Mixed_6h/Branch_3/Conv2d_0b_1x1", "features.17.branch3.1.conv"),
    # ReductionB (Mixed_7a)
    ("InceptionV4/Mixed_7a/Branch_0/Conv2d_0a_1x1", "features.18.branch0.0.conv"),
    ("InceptionV4/Mixed_7a/Branch_0/Conv2d_1a_3x3", "features.18.branch0.1.conv"),
    ("InceptionV4/Mixed_7a/Branch_1/Conv2d_0a_1x1", "features.18.branch1.0.conv"),
    ("InceptionV4/Mixed_7a/Branch_1/Conv2d_0b_1x7", "features.18.branch1.1.conv"),
    ("InceptionV4/Mixed_7a/Branch_1/Conv2d_0c_7x1", "features.18.branch1.2.conv"),
    ("InceptionV4/Mixed_7a/Branch_1/Conv2d_1a_3x3", "features.18.branch1.3.conv"),
    # InceptionC (1)
    ("InceptionV4/Mixed_7b/Branch_0/Conv2d_0a_1x1", "features.19.branch0.conv"),
    ("InceptionV4/Mixed_7b/Branch_1/Conv2d_0a_1x1", "features.19.branch1_0.conv"),
    ("InceptionV4/Mixed_7b/Branch_1/Conv2d_0b_1x3", "features.19.branch1_1a.conv"),
    ("InceptionV4/Mixed_7b/Branch_1/Conv2d_0c_3x1", "features.19.branch1_1b.conv"),
    ("InceptionV4/Mixed_7b/Branch_2/Conv2d_0a_1x1", "features.19.branch2_0.conv"),
    ("InceptionV4/Mixed_7b/Branch_2/Conv2d_0b_3x1", "features.19.branch2_1.conv"),
    ("InceptionV4/Mixed_7b/Branch_2/Conv2d_0c_1x3", "features.19.branch2_2.conv"),
    ("InceptionV4/Mixed_7b/Branch_2/Conv2d_0d_1x3", "features.19.branch2_3a.conv"),
    ("InceptionV4/Mixed_7b/Branch_2/Conv2d_0e_3x1", "features.19.branch2_3b.conv"),
    ("InceptionV4/Mixed_7b/Branch_3/Conv2d_0b_1x1", "features.19.branch3.1.conv"),
    # InceptionC (2)
    ("InceptionV4/Mixed_7c/Branch_0/Conv2d_0a_1x1", "features.20.branch0.conv"),
    ("InceptionV4/Mixed_7c/Branch_1/Conv2d_0a_1x1", "features.20.branch1_0.conv"),
    ("InceptionV4/Mixed_7c/Branch_1/Conv2d_0b_1x3", "features.20.branch1_1a.conv"),
    ("InceptionV4/Mixed_7c/Branch_1/Conv2d_0c_3x1", "features.20.branch1_1b.conv"),
    ("InceptionV4/Mixed_7c/Branch_2/Conv2d_0a_1x1", "features.20.branch2_0.conv"),
    ("InceptionV4/Mixed_7c/Branch_2/Conv2d_0b_3x1", "features.20.branch2_1.conv"),
    ("InceptionV4/Mixed_7c/Branch_2/Conv2d_0c_1x3", "features.20.branch2_2.conv"),
    ("InceptionV4/Mixed_7c/Branch_2/Conv2d_0d_1x3", "features.20.branch2_3a.conv"),
    ("InceptionV4/Mixed_7c/Branch_2/Conv2d_0e_3x1", "features.20.branch2_3b.conv"),
    ("InceptionV4/Mixed_7c/Branch_3/Conv2d_0b_1x1", "features.20.branch3.1.conv"),
    # InceptionC (3)
    ("InceptionV4/Mixed_7d/Branch_0/Conv2d_0a_1x1", "features.21.branch0.conv"),
    ("InceptionV4/Mixed_7d/Branch_1/Conv2d_0a_1x1", "features.21.branch1_0.conv"),
    ("InceptionV4/Mixed_7d/Branch_1/Conv2d_0b_1x3", "features.21.branch1_1a.conv"),
    ("InceptionV4/Mixed_7d/Branch_1/Conv2d_0c_3x1", "features.21.branch1_1b.conv"),
    ("InceptionV4/Mixed_7d/Branch_2/Conv2d_0a_1x1", "features.21.branch2_0.conv"),
    ("InceptionV4/Mixed_7d/Branch_2/Conv2d_0b_3x1", "features.21.branch2_1.conv"),
    ("InceptionV4/Mixed_7d/Branch_2/Conv2d_0c_1x3", "features.21.branch2_2.conv"),
    ("InceptionV4/Mixed_7d/Branch_2/Conv2d_0d_1x3", "features.21.branch2_3a.conv"),
    ("InceptionV4/Mixed_7d/Branch_2/Conv2d_0e_3x1", "features.21.branch2_3b.conv"),
    ("InceptionV4/Mixed_7d/Branch_3/Conv2d_0b_1x1", "features.21.branch3.1.conv"),
    # Fully-connected layer
    ("InceptionV4/Logits/Logits", "last_linear"),
    # There is also an auxiliary head but the pytorch impl does not include it.
]


def convert_tf_to_pytorch(input_path, output_path, num_classes: int):
    try:
        ckpt = tf.train.load_checkpoint(input_path)
    except tf.errors.DataLossError:
        raise RuntimeError(
            "Error: could not load checkpoint. Did you pass in the stem of the path?"
            "Pass in the path without '.index' or '.meta' or '.data-00000-of-00001'."
        )
    new_state_dict = {}
    for tf_prefix, torch_prefix in tf_to_pytorch_layers:
        tf_weights = f"{tf_prefix}/weights"
        tf_biases = f"{tf_prefix}/biases"
        torch_weights = f"{torch_prefix}.weight"
        torch_biases = f"{torch_prefix}.bias"
        tf_weight_array = ckpt.get_tensor(tf_weights)
        tf_bias_array = ckpt.get_tensor(tf_biases)
        if "/Conv2d" in tf_weights:
            # (1, 3, 512, 256) -> (256, 512, 1, 3)
            tf_weight_array = tf_weight_array.transpose([3, 2, 0, 1])
        elif tf_weights == "InceptionV4/Logits/Logits/weights":
            # (1536, 2) -> (2, 1536)
            tf_weight_array = tf_weight_array.transpose([1, 0])
        new_state_dict[torch_weights] = torch.from_numpy(tf_weight_array)
        new_state_dict[torch_biases] = torch.from_numpy(tf_bias_array)

    if inceptionv4 is None:
        print("Not testing model weights because inceptionv4 module not found.")
    else:
        # Test that conversion was (probably) done correctly.
        true_model = inceptionv4(num_classes=num_classes, pretrained=False)
        if true_model.state_dict().keys() != new_state_dict.keys():
            raise RuntimeError(
                "Something went wrong... converted model keys do not match InceptionV4"
                " keys in PyTorch."
            )
        true_state_dict = true_model.state_dict()
        for true_k, new_k in zip(true_state_dict, new_state_dict):
            true_shape = true_state_dict[true_k].shape
            new_shape = new_state_dict[new_k].shape
            if true_shape != new_shape:
                raise ValueError(
                    f"Shape mismatch in {true_k}: {true_shape} vs {new_shape}"
                )

    torch.save(new_state_dict, output_path)


if __name__ == "__main__":
    p = argparse.ArgumentParser(description=__doc__)
    p.add_argument("input", help="Path to the InceptionV4 tensorflow checkpoint.")
    p.add_argument("output")
    p.add_argument(
        "--num-classes", type=int, default=2, help="Number of output classes."
    )
    args = p.parse_args()
    # We do not expect the input "path" to exist, because it should be the stem.
    if Path(args.input).exists():
        raise ValueError(
            "Input path exists. Instead of passing the full path, pass in the stem"
            " (i.e., without '.index' or '.meta' or '.data-00000-of-00001'."
        )
    convert_tf_to_pytorch(args.input, args.output, args.num_classes)
